import io
import tarfile
from pathlib import Path
from typing import IO, TextIO, Union, AnyStr
from urllib.parse import urlparse

import requests

FilePathOrBuffer = Union[str, Path, IO[AnyStr], TextIO]
Buffer = Union[TextIO, IO[AnyStr]]


def get_handle_to_write(filepath_or_buffer: FilePathOrBuffer, mode: str = 'r', **kwargs) -> IO[AnyStr]:
    """
    Function that returns the file handler from the Path object or the buffer object or the url.

    Parameters
    ----------
    filepath_or_buffer : str | Path | IO[AnyStr] | TextIO
        file path

    mode : str, optional
        file mode, by default 'r'

    Returns
    -------
    handler : IO[AnyStr]
        file handler

    Raises
    ------
    ValueError
        if the filepath_or_buffer is not a valid file path or buffer or url
    """

    if is_buffer(filepath_or_buffer):
        return filepath_or_buffer

    elif is_url(filepath_or_buffer, **kwargs):
        raise ValueError("Cannot write to a url.")

    return open(filepath_or_buffer, mode=mode)


def is_url(filepath_or_buffer: FilePathOrBuffer, **kwargs) -> bool:
    """
    Function that checks if the filepath_or_buffer is a url.

    Parameters
    ----------
    filepath_or_buffer : str | Path | IO[AnyStr] | TextIO
        file path

    Returns
    -------
    is_url : bool
        True if the filepath_or_buffer is an url, False otherwise.
    """
    if not isinstance(filepath_or_buffer, str):
        return False

    if urlparse(filepath_or_buffer).scheme != '':
        try:

            response = requests.head(filepath_or_buffer, allow_redirects=True, **kwargs)
            if response.status_code == 200:
                return True

            return False

        except requests.exceptions.RequestException:
            return False

    return False


def is_buffer(filepath_or_buffer: FilePathOrBuffer) -> bool:
    """
    Function that checks if the filepath_or_buffer is a buffer.

    Parameters
    ----------
    filepath_or_buffer : str | Path | IO[AnyStr] | TextIO
        file path

    Returns
    -------
    is_buffer : bool
        True if the filepath_or_buffer is a buffer, False otherwise.
    """
    return isinstance(filepath_or_buffer, (io.TextIOBase,
                                           io.BufferedIOBase,
                                           io.RawIOBase,
                                           io.IOBase))


def is_path(filepath_or_buffer: FilePathOrBuffer) -> bool:
    """
    Function that checks if the filepath_or_buffer is a path.

    Parameters
    ----------
    filepath_or_buffer : str | Path | IO[AnyStr] | TextIO
        file path

    Returns
    -------
    is_path : bool
        True if the filepath_or_buffer is a path, False otherwise.
    """
    if isinstance(filepath_or_buffer, Path):
        return True

    path = Path(filepath_or_buffer)
    try:
        if path.exists():
            return True
        return False

    except OSError:
        return False


def is_dir(filepath_or_buffer: FilePathOrBuffer) -> bool:
    """
    Function that checks if the DirPath is a directory.

    Parameters
    ----------
    filepath_or_buffer : str | Path
        directory path path

    Returns
    -------
    is_path : booldir_path
        True if the dir_path is a directory, False otherwise.
    """
    if not isinstance(filepath_or_buffer, Path):
        filepath_or_buffer = Path(filepath_or_buffer)
    try:
        if filepath_or_buffer.is_dir():
            return True
        return False

    except OSError:
        return False


def is_gzip_tarball(filepath_or_buffer: FilePathOrBuffer) -> bool:
    """
    Function that checks if the FilePathOrBuffer is a gziped tarball.

    Parameters
    ----------
    filepath_or_buffer : str | Path | IO[AnyStr] | TextIO
        directory path path

    Returns
    -------
    is_path : bool
        True if the FilePathOrBuffer is a gziped tarball, False otherwise.
    """
    try:
        if tarfile.is_tarfile(filepath_or_buffer):
            return True
        return False

    except OSError:
        return False


def is_gzip_tarball_path(filepath_or_buffer: FilePathOrBuffer) -> bool:
    """
    Checks that the provided path is in good-faith a .tar.gz tarball.

    It only checks the path for situations where it is not possible to check the content directly, like in URL before
    downloading. This is a very unsafe assumption! As soon as you can, you must use the is_gzip_tarball() function to
    check the contents directly.

    Parameters
    ----------
    filepath_or_buffer : str | Path | IO[AnyStr] | TextIO
        directory path path

    Returns
    -------
    is_path : bool
        True if the FilePathOrBuffer is a gziped tarball, False otherwise.
    """
    suffix = ".tar.gz"

    if isinstance(filepath_or_buffer, Path):
        if filepath_or_buffer.name.endswith(suffix):
            return True
        return False
    try:
        if filepath_or_buffer.endswith(suffix):
            return True
        return False

    except OSError:
        return False


def get_handle_to_read(filepath_or_buffer: FilePathOrBuffer, mode: str = 'r', **kwargs) -> IO[AnyStr]:
    """
    Function that returns the file handler from the Path object or the buffer object or the url.

    Parameters
    ----------
    filepath_or_buffer : str | Path | IO[AnyStr] | TextIO
        file path

    mode : str, optional
        file mode, by default 'r'

    Returns
    -------
    handler : IO[AnyStr]
        file handler

    Raises
    ------
    ValueError
        if the filepath_or_buffer is not a valid file path or buffer or url
    """
    if is_buffer(filepath_or_buffer):
        return filepath_or_buffer

    elif is_path(filepath_or_buffer):
        return open(filepath_or_buffer, mode=mode)

    elif is_url(filepath_or_buffer, **kwargs):
        data = requests.get(filepath_or_buffer, **kwargs)

        if is_gzip_tarball_path(filepath_or_buffer):
            tar = tarfile.open(fileobj=io.BytesIO(data.content))
            return tar.extractfile(tar.getnames()[0])

        if mode == 'r':
            return io.StringIO(data.text)

        elif mode == 'rb':
            return io.BytesIO(data.content)

        raise ValueError("Cannot read from a url.")

    raise ValueError("No file path or buffer or url provided.")


def get_buffer(filepath_or_buffer: FilePathOrBuffer, mode: str = 'r', **kwargs) -> Buffer:
    """
    Function that opens the file and returns a Buffer object

    Parameters
    ----------
    filepath_or_buffer : str | Path | IO[AnyStr] | TextIO
        file path
    mode : str
        mode of buffering (e.g., 'w' for writing, 'r' for reading)

    Returns
    -------
    """
    if hasattr(filepath_or_buffer, 'buffer'):
        return filepath_or_buffer

    else:
        return open(filepath_or_buffer, mode, **kwargs)


def get_path(filepath_or_buffer: FilePathOrBuffer, **kwargs) -> Path:
    """
    Function that returns a Path object.

    Parameters
    ----------
    filepath_or_buffer : str | Path | IO[AnyStr] | TextIO
        file path

    Returns
    -------
    path : Path
    """
    if isinstance(filepath_or_buffer, TextIO) or isinstance(filepath_or_buffer, IO):
        return Path(filepath_or_buffer.name, **kwargs)

    return Path(filepath_or_buffer, **kwargs)


from collections import defaultdict, Counter
from statistics import mean

from Bio.SeqUtils.ProtParam import ProteinAnalysis
from modlamp.descriptors import GlobalDescriptor, PeptideDescriptor

from ._constants import BONDS
from ._utils import Descriptor


class LengthDescriptor(Descriptor):
    """
    A descriptor that returns the length of the sequence.
    """

    def __call__(self, sequence: str, **kwargs):
        return [float(len(sequence.strip()))]

    def get_features_out(self, **kwargs):
        return ["length"]


class ChargeDescriptor(Descriptor):
    """
    A descriptor that returns the charge of the sequence.
    """

    def __call__(self, sequence: str, ph: float = 7.4, amide: bool = False, **kwargs):
        """
        Calculates the charge of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the charge for.

        ph: float, default 7.4
            ph considered for the calculation.

        amide: bool, default False
            Considering as an amide protein sequence.
        kwargs

        Returns
        -------
        list of float
            The charge of the sequence.
        """
        desc = GlobalDescriptor(sequence)
        desc.calculate_charge(ph=ph, amide=amide)
        return [desc.descriptor[0][0]]

    def get_features_out(self, **kwargs):
        return ["charge"]


class ChargeDensityDescriptor(Descriptor):
    """
    A descriptor that returns the charge density of the sequence.
    """

    def __call__(self, sequence: str, ph: float = 7.4, amide: bool = False, **kwargs):
        """
        Calculates the charge density of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the charge density for.

        ph: float, default 7.4
            ph considered for the calculation.

        amide: bool, default False
            Considering as an amide protein sequence.
        kwargs

        Returns
        -------
        list of float
            The charge density of the sequence.
        """
        desc = GlobalDescriptor(sequence)
        desc.charge_density(ph=ph, amide=amide)
        return [desc.descriptor[0][0]]

    def get_features_out(self, **kwargs):
        return ["charge_density"]


class FormulaDescriptor(Descriptor):
    """
    A descriptor that returns the formula of the sequence.
    """

    def __call__(self, sequence: str, amide: bool = False, **kwargs):
        """
        Calculates the formula of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the formula for.

        amide: bool, default False
            Considering as an amide protein sequence.

        kwargs

        Returns
        -------
        list of float
            The formula of the sequence.
        """
        desc = GlobalDescriptor(sequence)
        desc.formula(amide)
        counts = desc.descriptor[0][0].split()
        counts = {atom[0]: int(atom[1:]) for atom in counts}

        formula = {'C': 0, 'H': 0, 'N': 0, 'O': 0, 'S': 0}
        formula.update(counts)
        return list(formula.values())

    def get_features_out(self, **kwargs):
        return ['Carbon', 'Hydrogen', 'Nitrogen', 'Oxygen', 'Sulfur']


class BondDescriptor(Descriptor):
    """
    A descriptor that returns the bond order of the sequence.
    """

    def __call__(self, sequence: str, **kwargs):
        """
        Calculates the bond order of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the bond order for.

        kwargs

        Returns
        -------
        list of float
            The bond order of the sequence.
        """
        aa_counts = Counter(sequence.upper())

        bonds = defaultdict(int)
        for aa, cnt in aa_counts.items():
            if aa in BONDS:
                bonds['total'] += BONDS[aa]['total_bounds'] * cnt
                bonds['hydrogen'] += BONDS[aa]['hydrogen_bonds'] * cnt
                bonds['single'] += BONDS[aa]['single_bounds'] * cnt
                bonds['double'] += BONDS[aa]['double_bounds'] * cnt
        return list(bonds.values())

    def get_features_out(self, **kwargs):
        return ['total', 'hydrogen', 'single', 'double']


class MolecularWeightDescriptor(Descriptor):
    """
    A descriptor that returns the molecular weight of the sequence.
    """

    def __call__(self, sequence: str, **kwargs):
        """
        Calculates the molecular weight of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the molecular weight for.

        kwargs

        Returns
        -------
        list of float
            The molecular weight of the sequence.
        """
        desc = GlobalDescriptor(sequence)
        desc.calculate_MW(amide=True)
        return [desc.descriptor[0][0]]

    def get_features_out(self, **kwargs):
        return ["molecular_weight"]


class GravyDescriptor(Descriptor):
    """
    A descriptor that returns the gravy of the sequence.
    """

    def __call__(self, sequence: str, **kwargs):
        """
        Calculates the gravy of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the gravy for.

        kwargs

        Returns
        -------
        list of float
            The gravy of the sequence.
        """
        analysis = ProteinAnalysis(sequence)
        return [analysis.gravy()]

    def get_features_out(self, **kwargs):
        return ["gravy"]


class AromacityDescriptor(Descriptor):
    """
    A descriptor that returns the aromacity of the sequence.
    """

    def __call__(self, sequence: str, **kwargs):
        """
        Calculates the aromacity of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the aromacity for.

        kwargs

        Returns
        -------
        list of float
            The aromacity of the sequence.
        """
        analysis = ProteinAnalysis(sequence)
        return [analysis.aromaticity()]

    def get_features_out(self, **kwargs):
        return ["aromaticity"]


class IsoelectricPointDescriptor(Descriptor):
    """
    A descriptor that returns the isoelectric point of the sequence.
    """

    def __call__(self, sequence: str, **kwargs):
        """
        Calculates the isoelectric point of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the isoelectric point for.

        kwargs

        Returns
        -------
        list of float
            The isoelectric point of the sequence.
        """
        analysis = ProteinAnalysis(sequence)
        return [analysis.isoelectric_point()]

    def get_features_out(self, **kwargs):
        return ["isoelectric_point"]


class InstabilityIndexDescriptor(Descriptor):
    """
    A descriptor that returns the instability index of the sequence.
    """

    def __call__(self, sequence: str, **kwargs):
        """
        Calculates the instability index of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the instability index for.

        kwargs

        Returns
        -------
        list of float
            The instability index of the sequence.
        """
        analysis = ProteinAnalysis(sequence)
        return [analysis.instability_index()]

    def get_features_out(self, **kwargs):
        return ["instability_index"]


class SecondaryStructureDescriptor(Descriptor):
    """
    A descriptor that returns the secondary structure of the sequence.
    """

    def __call__(self, sequence: str, **kwargs):
        """
        Calculates the secondary structure of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the secondary structure for.

        kwargs

        Returns
        -------
        list of float
            The secondary structure of the sequence.
        """
        analysis = ProteinAnalysis(sequence)
        helix, turn, sheet = analysis.secondary_structure_fraction()
        return [helix, turn, sheet]

    def get_features_out(self, **kwargs):
        return ["secondary_structure_helix", "secondary_structure_turn", "secondary_structure_sheet"]


class MolarExtinctionCoefficientDescriptor(Descriptor):
    """
    A descriptor that returns the molar extinction coefficient of the sequence.
    """

    def __call__(self, sequence: str, **kwargs):
        """
        Calculates the molar extinction coefficient of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the molar extinction coefficient for.

        kwargs

        Returns
        -------
        list of float
            The molar extinction coefficient of the sequence.
        """
        analysis = ProteinAnalysis(sequence)
        reduced, oxidized = analysis.molar_extinction_coefficient()
        return [reduced, oxidized]

    def get_features_out(self, **kwargs):
        return ["molar_extinction_coefficient_reduced", "molar_extinction_coefficient_oxidized"]


class FlexibilityDescriptor(Descriptor):
    """
    A descriptor that returns the flexibility of the sequence.
    """

    def __call__(self, sequence: str, **kwargs):
        """
        Calculates the flexibility of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the flexibility for.

        kwargs

        Returns
        -------
        list of float
            The flexibility of the sequence.
        """
        # TODO: it should be confirmed if this is the correct way to calculate the flexibility.
        #  The original code was returning the list of all the flexibility values.
        #  However, the name of the feature cannot vary according to the sequence length.
        analysis = ProteinAnalysis(sequence)
        flexibility = analysis.flexibility()
        return [mean(flexibility)]

    def get_features_out(self, **kwargs):
        return ["flexibility"]


class AliphaticIndexDescriptor(Descriptor):
    """
    A descriptor that returns the aliphatic index of the sequence.
    """

    def __call__(self, sequence: str, **kwargs):
        """
        Calculates the aliphatic index of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the aliphatic index for.

        kwargs

        Returns
        -------
        list of float
            The aliphatic index of the sequence.
        """
        desc = GlobalDescriptor(sequence)
        desc.aliphatic_index()
        return [desc.descriptor[0][0]]

    def get_features_out(self, **kwargs):
        return ["aliphatic_index"]


class BomanIndexDescriptor(Descriptor):
    """
    A descriptor that returns the Boman index of the sequence.
    """

    def __call__(self, sequence: str, **kwargs):
        """
        Calculates the Boman index of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the Boman index for.

        kwargs

        Returns
        -------
        list of float
            The Boman index of the sequence.
        """
        desc = GlobalDescriptor(sequence)
        desc.boman_index()
        return [desc.descriptor[0][0]]

    def get_features_out(self, **kwargs):
        return ["boman_index"]


class HydrophobicRatioDescriptor(Descriptor):
    """
    A descriptor that returns the hydrophobic ratio of the sequence.
    """

    def __call__(self, sequence: str, **kwargs):
        """
        Calculates the hydrophobic ratio of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the hydrophobic ratio for.

        kwargs

        Returns
        -------
        list of float
            The hydrophobic ratio of the sequence.
        """
        desc = GlobalDescriptor(sequence)
        desc.hydrophobic_ratio()
        return [desc.descriptor[0][0]]

    def get_features_out(self, **kwargs):
        return ["hydrophobic_ratio"]


class ModlAMPMomentDescriptor(Descriptor):
    """
    A descriptor that returns the moment of the sequence.
    """

    def __call__(self, sequence: str,
                 window: int = 7,
                 scalename: str = 'Eisenberg',
                 angle: int = 100,
                 modality: str = 'max',
                 **kwargs):
        """
        Calculates the moment of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the moment for.

        scalename : str
            The name of the scale to use.

        angle : int
            The angle to use for the moment calculation.

        modality : str
            The modality to use for the moment calculation.

        kwargs

        Returns
        -------
        list of float
            The moment of the sequence.
        """
        desc = PeptideDescriptor(sequence, scalename)
        desc.calculate_moment(window, angle, modality)
        return [desc.descriptor[0][0]]

    def get_features_out(self, **kwargs):
        return ["moment"]


class ModlAMPGlobalDescriptor(Descriptor):
    """
    A descriptor that returns the global of the sequence.
    """

    def __call__(self, sequence: str,
                 window: int = 7,
                 scalename: str = 'Eisenberg',
                 modality: str = 'max',
                 **kwargs):
        """
        Calculates the global of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the global for.

        scalename : str
            The name of the scale to use.

        modality : str
            The modality to use for the global calculation.

        kwargs

        Returns
        -------
        list of float
            The global of the sequence.
        """
        desc = PeptideDescriptor(sequence, scalename)
        desc.calculate_global(window, modality)
        return [desc.descriptor[0][0]]

    def get_features_out(self, **kwargs):
        return ["global"]


class ModlAMPProfileDescriptor(Descriptor):
    """
    A descriptor that returns the profile of the sequence.
    """
    def __call__(self, sequence: str,
                 window: int = 7,
                 scalename: str = 'Eisenberg',
                 prof_type: str = 'uH',
                 **kwargs):
        """
        Calculates the profile of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the profile for.

        scalename : str
            The name of the scale to use.

        prof_type : str
            The type of profile to calculate.

        kwargs

        Returns
        -------
        list of float
            The profile of the sequence.
        """
        desc = PeptideDescriptor(sequence, scalename)
        desc.calculate_profile(prof_type, window)
        return list(desc.descriptor[0])

    def get_features_out(self, **kwargs):
        # profile descriptor returns only two values
        return ['profile_1', 'profile_2']


class ModlAMPArcDescriptor(Descriptor):
    """
    A descriptor that returns the arcs of the sequence.
    """
    def __call__(self, sequence: str,
                 window: int = 7,
                 scalename: str = 'peparc',
                 modality: str = 'max',
                 **kwargs):
        """
        Calculates the arcs of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the arcs for.

        scalename : str
            The name of the scale to use.

        modality : str
            The modality to use for the global calculation.

        kwargs

        Returns
        -------
        list of float
            The arcs of the sequence.
        """
        desc = PeptideDescriptor(sequence, scalename)
        desc.calculate_arc(modality)
        return list(desc.descriptor[0])

    def get_features_out(self, **kwargs):
        # arc descriptor returns only five values
        return [f'arc_{i}' for i in range(1, 5 + 1)]



from typing import Any

import numpy as np

from plants_sm.data_structures.dataset import Dataset
from plants_sm.featurization.featurizer import FeaturesGenerator
from plants_sm.featurization.proteins.encodings._constants import BLOSUM_62, BLOSUM_50


class BLOSSUMEncoder(FeaturesGenerator):
    """
    It encodes protein sequences with a given BLOSUM substitution matrix.

    The number 62 refers to the percentage identity at which sequences are clustered in the analysis.
    It is possible to get 50 to get 50 % identity.
    Encoding a peptide this way means we provide the column from the blosum matrix corresponding to the amino acid
    at each position of the sequence. This produces 24 * seq_len matrix.

    The ProteinBlosum operates only over pandas DataFrame. Depending on the blosum used, it may require different
    preprocessing of protein sequences. For blosum62, protein sequences can not include the amino acid such as J, U, O.
    For blosum50, protein sequences can not include the amino acid such as U, O.

    """

    blosum: str = "blosum62"
    _blosum: dict = None
    output_shape_dimension = 3
    name = f"blosum_{blosum}"

    def set_features_names(self):
        """
        It sets the features names.
        """
        self.features_names = [f"{self.blosum}_{num}" for num in range(1, len(self._blosum["A"]) + 1)]

    def _fit(self, dataset: Dataset, instance_type: str) -> 'BLOSSUMEncoder':
        """
        It fits the blosumEncoder.

        Parameters
        ----------
        dataset : Dataset
            Dataset containing the protein sequences to be encoded.

        Returns
        -------
        blosumEncoder

        """
        if self.blosum == "blosum62":
            self._blosum = BLOSUM_62

        elif self.blosum == "blosum50":
            self._blosum = BLOSUM_50

        else:
            ValueError(f'{self.blosum} is not a valid blosum.')

        return self

    def _fit_batch(self, dataset: Dataset, instance_type: str) -> 'BLOSSUMEncoder':
        """
        It fits the blosumEncoder.

        Parameters
        ----------
        dataset : Dataset
            Dataset containing the protein sequences to be encoded.

        Returns
        -------
        blosumEncoder

        """
        if self.blosum == "blosum62":
            self._blosum = BLOSUM_62

        elif self.blosum == "blosum50":
            self._blosum = BLOSUM_50

        else:
            ValueError(f'{self.blosum} is not a valid blosum.')

        return self

    def _featurize(self, sequence: Any) -> np.ndarray:

        return np.array([np.array(self._blosum[aa]) for aa in sequence], dtype=np.float32)



from abc import ABCMeta, abstractmethod
from contextlib import contextmanager
from typing import List, Any

from plants_sm.io.commons import FilePathOrBuffer, get_path, get_buffer


class Writer(metaclass=ABCMeta):
    """
    Generic class for writing data from file(s).
    All writers must implement the read method.
    All writers must define the file_types attribute.
    """

    def __init__(self, file_path_or_buffer: FilePathOrBuffer, **kwargs):
        """
        Initializer of this class that defines instance variables such as the buffer for the file and path.

        Parameters
        ----------
        file_path_or_buffer : str | Path | IO[AnyStr] | TextIO
            file path

        Returns
        -------
        """

        self.path = get_path(file_path_or_buffer)
        if "mode" in kwargs:
            mode = kwargs.pop("mode")
        else:
            mode = "w"
        self.buffer = get_buffer(file_path_or_buffer, mode=mode)
        self.kwargs = kwargs

    def close_buffer(self):
        """
        Method to close buffer.
        """
        self.buffer.close()

    @staticmethod
    @abstractmethod
    def file_types() -> List[str]:
        """
        Abstract method and property that returns the file types that the writer can write.

        Returns
        -------
        file_types : List[str]
            file types that the writer can write
        """

    @contextmanager
    @abstractmethod
    def write(self, object_to_be_written: Any) -> bool:
        """
        Abstract method to read the data from file(s).

        Parameters
        ----------
        object_to_be_written : Any
            object to be written (e.g., dataframe or configuration file)

        Returns
        -------
        bool : True if the data was written successfully, False otherwise.
        """

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_buffer()
        return False



from abc import abstractmethod
import os
from typing import Dict, Union
from plants_sm.models.model import Model
from plants_sm.data_structures.dataset.dataset import Dataset
from plants_sm.models.constants import FileConstants
from plants_sm.io.pickle import read_pickle, write_pickle
from torch.utils.data import TensorDataset, DataLoader
import lightning as L
import torch
from plants_sm.models._utils import _get_pred_from_proba, _convert_proba_to_unified_form, write_model_parameters_to_pickle
from plants_sm.models.constants import BINARY, FileConstants

from plants_sm.models._utils import _convert_proba_to_unified_form, \
    write_model_parameters_to_pickle

import numpy as np

class InternalLightningModel(Model):

    def __init__(self, module, batch_size: int = 32, devices="cpu", **trainer_kwargs) -> None:
        super().__init__()

        self.module = module
        self.batch_size = batch_size
        self.trainer_kwargs = trainer_kwargs
        if isinstance(devices, list):
            self.ddp = True
            self.trainer = L.Trainer(devices=devices, **trainer_kwargs)
        else:
            self.ddp = False
            self.trainer = L.Trainer(**trainer_kwargs)

        self.devices = devices

    def get_embedding(self, dataset):
        if isinstance(dataset, Dataset):
            predict_dataloader = self._preprocess_data(dataset, shuffle=False)
        else:
            predict_dataloader = dataset

        # the module has to have the return_embedding attribute
        self.module.return_embedding = True 
        embeddings = self.trainer.predict(self.module, predict_dataloader)
        # get a list of embeddings from the list of tuples
        embeddings = [embedding[1] for embedding in embeddings]
        embeddings = torch.cat(embeddings)
        return np.array(embeddings)


    def _preprocess_data(self, dataset: Dataset, shuffle: bool = True) -> Dataset:
        """
        Preprocesses the data.

        Parameters
        ----------
        dataset: Dataset
            The dataset to preprocess.
        shuffle: bool
            Whether to shuffle the data

        Returns
        -------
        Dataset
            The preprocessed dataset.
        """
        tensors = []
        if isinstance(dataset.X, Dict):
            for instance in dataset.X.keys():
                tensor = torch.tensor(dataset.X[instance], dtype=torch.float)
                tensors.append(tensor)
        else:
            tensor = torch.tensor(dataset.X, dtype=torch.float)
            tensors.append(tensor)

        try:
            if dataset.y is not None:
                tensors.append(torch.tensor(dataset.y, dtype=torch.float))
        except ValueError:
            pass

        dataset = TensorDataset(
            *tensors
        )

        data_loader = DataLoader(
            dataset,
            shuffle=shuffle,
            batch_size=self.batch_size
        )
        return data_loader
    
    def _fit_data(self, train_dataset: Union[Dataset, TensorDataset], 
                  validation_dataset: Union[Dataset, TensorDataset]):
        """
        Fits the model to the data.

        Parameters
        ----------
        train_dataset: Dataset
            The dataset to fit the model to.
        validation_dataset: Dataset
            The dataset to validate the model on.
        """

        self.trainer = L.Trainer(devices=self.devices, **self.trainer_kwargs)
        if isinstance(train_dataset, Dataset):
            train_dataset_loader = self._preprocess_data(train_dataset)
        else:
            train_dataset_loader = train_dataset

        if validation_dataset is not None:
            if isinstance(train_dataset, Dataset):
                validation_dataloader = self._preprocess_data(validation_dataset, shuffle=False)
            else:
                validation_dataloader = validation_dataset
        if validation_dataset:
            self.trainer.fit(model=self.module, train_dataloaders=train_dataset_loader, val_dataloaders=validation_dataloader)
        else:
            self.trainer.fit(model=self.module, train_dataloaders=train_dataset_loader)

    def _predict_proba(self, dataset: Union[Dataset, TensorDataset], trainer = L.Trainer(accelerator="cpu")) -> Union[np.ndarray, list]:
        """
        Predicts the probabilities of the classes.

        Parameters
        ----------
        dataset: Dataset
            The dataset to predict the probabilities on.

        Returns
        -------
        Union[np.ndarray, list]
            np.ndarray if the network only has one output, list of np.ndarray if the network has multiple outputs.
        """
        if isinstance(dataset, Dataset):
            predict_dataloader = self._preprocess_data(dataset, shuffle=False)
        else:
            predict_dataloader = dataset

        predictions = trainer.predict(self.module, predict_dataloader)
        if type(predictions[0]) == tuple:
            len_tuple = len(predictions[0])
            new_predictions = [None] * len_tuple
            for i in range(len_tuple):
                new_predictions[i] = [prediction[i] for prediction in predictions]
                new_predictions[i] = torch.cat(new_predictions[i]).detach().cpu().numpy()
            predictions = new_predictions
        else:
            predictions = torch.cat(predictions)
            # convert to numpy array
            predictions = predictions.detach().cpu().numpy()
        return predictions


    def _predict(self, dataset: Union[Dataset, TensorDataset], trainer = L.Trainer(accelerator="cpu")) -> np.ndarray:
        """
        Predicts the classes.

        Parameters
        ----------
        dataset: Dataset
            The dataset to predict the classes on.

        Returns
        -------
        np.ndarray
            The predicted classes.
        """
        predictions = self._predict_proba(dataset, trainer)

        if type(predictions) == list:
            for i in range(len(predictions)):
                predictions[i] = _convert_proba_to_unified_form(self.module.problem_type, np.array(predictions[i]))
                predictions[i] = _get_pred_from_proba(self.module.problem_type, predictions[i])
        else:
            predictions = _convert_proba_to_unified_form(self.module.problem_type, np.array(predictions))
            predictions = _get_pred_from_proba(self.module.problem_type, predictions)

        return predictions
    
    @classmethod
    def _load(cls, path: str):
        """
        Loads the model from a file.

        Parameters
        ----------
        path: str
            The path to load the model from.
        """
        weights_path = os.path.join(path, FileConstants.LIGHTNING_WEIGHTS.value)
        model = read_pickle(os.path.join(path, FileConstants.LIGHTNING_MODEL_PKL.value))
        model_parameters = read_pickle(os.path.join(path, FileConstants.MODEL_PARAMETERS_PKL.value))
        model = model.load_from_checkpoint(weights_path,**model_parameters)
        model = cls(module=model)
        model.trainer = read_pickle(os.path.join(path, "trainer.pk"))
        return model

    def _save(self, path: str):
        """
        Saves the model to a file.

        Parameters
        ----------
        path: str
            The path to save the model to.
        """
        weights_path = os.path.join(path, FileConstants.LIGHTNING_WEIGHTS.value)
        self.trainer.save_checkpoint(weights_path)
        write_pickle(os.path.join(path, "trainer.pk"), self.trainer)
        write_pickle(os.path.join(path, FileConstants.LIGHTNING_MODEL_PKL.value), self.module.__class__)
        write_model_parameters_to_pickle(self.module._contructor_parameters, path)

    @property
    def history(self):
        """
        Returns the underlying model.
        """


class InternalLightningModule(L.LightningModule):

    def __init__(self, problem_type: str = BINARY, metric = None):
        """
        Initializes the model.

        Parameters
        ----------
        batch_size: int
            The batch size to use.
        """
        super().__init__()
        self.problem_type = problem_type
        self.metric = metric

        self._contructor_parameters = { "problem_type": problem_type }

        self.training_step_outputs = []
        self.validation_step_outputs = []
        self.training_step_y_true = []
        self.validation_step_y_true = []
        self.epoch_losses = []
        self._update_constructor_parameters()

    @abstractmethod
    def _update_constructor_parameters():
        pass

    @abstractmethod
    def compute_loss(self, logits, y):
        pass
        
    def training_step(self, batch, batch_idx):
        x, y = batch
        if not isinstance(x, list):
            x = [x]
        logits = self(x)
        loss = self.compute_loss(logits, y)
        logits = logits.detach().cpu()
        y = y.detach().cpu()
        
        self.training_step_outputs.append(logits)
        self.training_step_y_true.append(y)
        self.log("train_loss", loss.item(), on_epoch=True, 
                 prog_bar=True, logger=True, sync_dist=True)
        return loss
    
    def validation_step(self, batch, batch_idx):
        inputs, target = batch
        if not isinstance(inputs, list):
            inputs = [inputs]
        output = self(inputs)

        output = output.detach().cpu()
        target = target.detach().cpu()

        self.validation_step_outputs.append(output)
        self.validation_step_y_true.append(target)

    def on_train_epoch_end(self) -> None:

        if self.metric is not None:

            predictions = _convert_proba_to_unified_form(self.problem_type, torch.cat(self.training_step_outputs).detach().cpu().numpy())
            predictions = _get_pred_from_proba(self.problem_type, predictions)
            self.log("train_metric", self.metric(torch.cat(self.training_step_y_true).detach().cpu().numpy(), predictions), 
                     on_epoch=True, prog_bar=True, logger=True, sync_dist=True)
        
        self.training_step_outputs = []
        self.training_step_y_true = []
    
    def on_validation_epoch_end(self) -> None:

        loss = self.compute_loss(torch.cat(self.validation_step_outputs), 
                                 torch.cat(self.validation_step_y_true))
        self.log("val_loss", loss, on_epoch=True, prog_bar=True, logger=True, sync_dist=True)

        if self.metric is not None:
            predictions = _convert_proba_to_unified_form(self.problem_type, torch.cat(self.validation_step_outputs).detach().cpu().numpy())
            predictions = _get_pred_from_proba(self.problem_type, predictions)
            self.log("val_metric", self.metric(torch.cat(self.validation_step_y_true).detach().cpu().numpy(), predictions),
                    on_epoch=True, prog_bar=True, logger=True, sync_dist=True)

        self.validation_step_outputs = []
        self.validation_step_y_true = []

    
    def predict_step(self, batch):
        if len(batch) == 2:
            inputs, target = batch
        else:
            inputs = batch
        if not isinstance(inputs, list):
            inputs = [inputs]
        return self(inputs)
    


from typing import Tuple

from plants_sm.data_structures.dataset import Dataset
from plants_sm.transformation._utils import transform_instances
from plants_sm.transformation.transformer import Transformer


class ProteinStandardizer(Transformer):
    """
    It transforms the protein sequences by replacing the following aminoacids
    by the closest amino acid residue available in the sequence:
        - Asparagine (B) -> Asparagine (N)
        - Glutamine(Z) -> Glutamine (Q)
        - Selenocysteine (U) -> Cysteinen(C)
        - Pyrrolysine (O) -> Lysine (K)

    It also removes the ambiguous amino acid (X) and replaces the ambiguous amino acid (J) by Isoleucine (I).

    Parameters
    ----------
    B: str, optional (default='N')
        One-letter aminoacid code to replace Asparagine (B). Default amino acid is Asparagine (N)
    Z   : str, optional (default='Q')
        One-letter aminoacid code to replace Glutamine(Z). Default amino acid is Glutamine (Q)
    U: str, optional (default='C')
        One-letter aminoacid code to replace Selenocysteine (U). Default amino acid is Cysteine (C)
    O: str, optional (default='K')
        One-letter aminoacid code to replace Pyrrolysine (O). Default amino acid is Lysine (K)
    J: str, optional (default='I')
        One-letter aminoacid code to replace ambiguous aminoacid (J). Default amino acid is Isoleucine (I)
    X: str, optional (default='')
        One-letter aminoacid code to remove ambiguous aminoacid (X). Default amino acid is removed
    """
    B: str = 'N'
    Z: str = 'Q'
    U: str = 'C'
    O: str = 'K'
    J: str = 'I'  # It can be either 'I' or 'L'
    X: str = ''

    def _fit(self, dataset: Dataset, instance_type: str) -> 'ProteinStandardizer':
        """
        Fit the transformer.

        Parameters
        ----------
        dataset : Dataset
            Dataset to fit.

        Returns
        -------
        self : ProteinStandardizer
            Fitted transformer.
        """
        self.replace_dict = {'B': self.B, 'Z': self.Z, 'U': self.U, 'O': self.O, 'J': self.J, 'X': self.X}
        return self

    def _fit_batch(self, dataset: Dataset, instance_type: str) -> 'ProteinStandardizer':
        """
        Fit the transformer.

        Parameters
        ----------
        dataset : Dataset
            Dataset to fit.

        Returns
        -------
        self : ProteinStandardizer
            Fitted transformer.
        """
        return self._fit(dataset, instance_type)

    def _transform(self, dataset: Dataset, instance_type: str) -> Dataset:
        """
        Standardize protein sequences.

        Parameters
        ----------
        dataset : Dataset
            Dataset to transform.

        Returns
        -------
        dataset : Dataset
            Transformed dataset.
        """
        return transform_instances(self.n_jobs, dataset, self._protein_preprocessing, instance_type,
                                   self.__class__.__name__)

    def _protein_preprocessing(self, sequence: str, identifier: str) -> Tuple[str, str]:
        """
        Preprocess a protein sequence.

        Parameters
        ----------
        sequence : str
            Protein sequence to preprocess.
        identifier : str
            Protein identifier.

        Returns
        -------
        Tuple[str, str]
            Dictionary with the identifier as key and the preprocessed sequence as value.
        """
        for key, value in self.replace_dict.items():
            sequence = sequence.replace(key, value)
        return identifier, sequence





from functools import wraps


def call_set_features_names(func):
    """
    Decorator that calls the method set_features_names before the method transform

    Parameters
    ----------
    func: _transform function always
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            if not self.features_names:
                self.set_features_names()
            res = func(self, *args, **kwargs)
            return res

        except ValueError as e:
            raise e

    return wrapper



import logging
from logging.handlers import TimedRotatingFileHandler

formatter = logging.Formatter('%(asctime)s %(name)s %(levelname)s %(message)s')
handler = TimedRotatingFileHandler('plants_sm', when='midnight', backupCount=20)
handler.setFormatter(formatter)
logger = logging.getLogger(__name__)
logger.addHandler(handler)
logger.setLevel(logging.DEBUG)


import torch
from sklearn.metrics import balanced_accuracy_score
from tensorflow.keras import Input
from tensorflow.keras.models import Model
from torch import nn, relu
from transformers import Conv1D
from tensorflow.keras.layers import Conv1D, concatenate, Embedding, GlobalMaxPooling1D, Dropout, Dense
import tensorflow as tf


class BalAccScore(tf.keras.callbacks.Callback):

    def __init__(self, X, y):
        super(BalAccScore, self).__init__()
        self.X = X
        self.y = y

    def on_train_begin(self, logs={}):
        self.balanced_accuracy = []

    def on_epoch_end(self, epoch, logs={}):
        y_predict = tf.argmax(self.model.predict(self.X), axis=1)
        y_true = tf.argmax(self.y, axis=1)
        balacc = balanced_accuracy_score(y_true.numpy(), y_predict.numpy())
        self.balanced_accuracy.append(round(balacc, 6))
        logs["val_bal_acc"] = balacc
        keys = list(logs.keys())

        print("\n ------ validation balanced accuracy score: %f ------\n" % balacc)


def balanced_accuracy(y_true, y_pred):
    from tensorflow.keras import backend as K
    def recall(y_true, y_pred):
        true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
        possible_positives = K.sum(K.round(K.clip(y_true, 0, 1)))
        recall_keras = true_positives / (possible_positives + K.epsilon())
        return recall_keras

    def specificity(y_true, y_pred):
        tn = K.sum(K.round(K.clip((1 - y_true) * (1 - y_pred), 0, 1)))
        fp = K.sum(K.round(K.clip((1 - y_true) * y_pred, 0, 1)))
        return tn / (tn + fp + K.epsilon())

    balanced_accuracy = (recall(y_true, y_pred) + specificity(y_true, y_pred)) / 2

    return balanced_accuracy


class DeepDTATensorflow:

    def __init__(self, PROT_LEN, SMILE_LEN, CHAR_PROT_SET_SIZE, CHAR_SMI_SET_SIZE, NUM_FILTERS):

        self.PROT_LEN = PROT_LEN
        self.SMILE_LEN = SMILE_LEN
        self.CHAR_SMI_SET_SIZE = CHAR_SMI_SET_SIZE
        self.CHAR_PROT_SET_SIZE = CHAR_PROT_SET_SIZE
        self.NUM_FILTERS = NUM_FILTERS
        self.model = self.build_model()

    def build_model(self):
        X_PROT = Input(shape=(self.PROT_LEN,), dtype='int32')
        X_MET = Input(shape=(self.SMILE_LEN,), dtype='int32')

        encode_protein = Embedding(input_dim=self.CHAR_PROT_SET_SIZE, output_dim=128, input_length=self.PROT_LEN)(
            X_PROT)
        encode_protein = Conv1D(filters=self.NUM_FILTERS, kernel_size=4, activation='relu', padding='valid', strides=1)(
            encode_protein)
        encode_protein = Conv1D(filters=self.NUM_FILTERS * 2, kernel_size=8, activation='relu', padding='valid',
                                strides=1)(
            encode_protein)
        encode_protein = Conv1D(filters=self.NUM_FILTERS * 3, kernel_size=10, activation='relu', padding='valid',
                                strides=1)(encode_protein)
        encode_protein = GlobalMaxPooling1D()(encode_protein)

        encode_smiles = Embedding(input_dim=self.CHAR_SMI_SET_SIZE, output_dim=128, input_length=self.SMILE_LEN)(X_MET)
        encode_smiles = Conv1D(filters=self.NUM_FILTERS, kernel_size=4, activation='relu', padding='valid', strides=1)(
            encode_smiles)
        encode_smiles = Conv1D(filters=self.NUM_FILTERS * 2, kernel_size=6, activation='relu', padding='valid',
                               strides=1)(
            encode_smiles)
        encode_smiles = Conv1D(filters=self.NUM_FILTERS * 3, kernel_size=8, activation='relu', padding='valid',
                               strides=1)(
            encode_smiles)
        encode_smiles = GlobalMaxPooling1D()(encode_smiles)

        encode_interaction = concatenate([encode_smiles, encode_protein],
                                         axis=-1)

        # Fully connected
        FC1 = Dense(1024, activation='relu')(encode_interaction)
        FC2 = Dropout(0.1)(FC1)
        FC2 = Dense(1024, activation='relu')(FC2)
        FC2 = Dropout(0.1)(FC2)
        FC2 = Dense(512, activation='relu')(FC2)

        # And add a logistic regression on top
        predictions = Dense(1, activation='sigmoid')(FC2)  # kernel_initializer='normal"

        interactionModel = Model(inputs=[X_PROT, X_MET], outputs=[predictions])

        interactionModel.compile(optimizer='adam', loss='binary_crossentropy', metrics=[balanced_accuracy])
        return interactionModel


class DeepDTA(nn.Module):

    def __init__(self, protein_shape, compounds_shape, protein_char_set_n, compound_char_set_n, filters):
        super().__init__()
        self.proteins_embedding = nn.Embedding(num_embeddings=protein_char_set_n + 1, embedding_dim=128, padding_idx=0)
        self.compounds_embedding = nn.Embedding(num_embeddings=compound_char_set_n + 1, embedding_dim=128,
                                                padding_idx=0)
        self.conv1_proteins_1 = nn.Conv1d(protein_shape[1], filters, 4, stride=1,
                                          padding='valid')
        self.conv1_proteins_2 = nn.Conv1d(filters, filters * 2, 6, stride=1, padding='valid')
        self.conv1_proteins_3 = nn.Conv1d(filters * 2, filters * 3, 8, stride=1, padding='valid')
        self.maxpool1_proteins = nn.MaxPool1d(2)

        self.conv1_compounds_1 = nn.Conv1d(compounds_shape[1], filters, 4, stride=1, padding='valid')
        self.conv1_compounds_2 = nn.Conv1d(filters, filters * 2, 6, stride=1, padding='valid')
        self.conv1_compounds_3 = nn.Conv1d(filters * 2, filters * 3, 8,
                                           stride=1, padding='valid')
        self.maxpool1_compounds = nn.MaxPool1d(2)

        self.dense1_interaction = nn.Linear(filters * 3 + filters * 3, 1024)
        self.dense2_interaction = nn.Linear(1024, 1024)
        self.dense3_interaction = nn.Linear(1024, 512)
        self.dropout = nn.Dropout(0.1)
        self.final_layer = nn.Linear(512, 1)

    def forward(self, x):
        x_proteins = x[0]
        x_proteins = x_proteins.to(torch.int32)
        x_proteins = self.proteins_embedding(x_proteins)
        y = relu(self.conv1_proteins_1(x_proteins))
        y = relu(self.conv1_proteins_2(y))
        y_proteins = relu(self.conv1_proteins_3(y))
        y_proteins = nn.MaxPool1d(y_proteins.shape[2] - 1)(y_proteins)

        x_compounds = x[1]
        x_compounds = x_compounds.to(torch.int32)
        x_compounds = self.compounds_embedding(x_compounds)
        y = relu(self.conv1_compounds_1(x_compounds))
        y = relu(self.conv1_compounds_2(y))
        y_compounds = relu(self.conv1_compounds_3(y))
        y_compounds = nn.MaxPool1d(y_compounds.shape[2] - 1)(y_compounds)

        y = torch.cat([y_proteins, y_compounds], dim=1)
        y = y.reshape(y.shape[0], y.shape[1])
        y = relu(self.dense1_interaction(y))
        y = self.dropout(y)
        y = relu(self.dense2_interaction(y))
        y = self.dropout(y)
        y = relu(self.dense3_interaction(y))
        y = self.dropout(y)
        y = torch.sigmoid(self.final_layer(y))
        return y



from typing import List

from ._constants import DISTANCE_1_POW_2, DISTANCE_2_POW_2, AA_ALPHABET
from ._utils import Descriptor, pairwise
from .amino_acid_composition import AminoAcidCompositionDescriptor


def sequence_order_coupling_number(protein_sequence: str,
                                   max_lag: int = 45,
                                   distance_matrix: dict = DISTANCE_1_POW_2) -> List[float]:
    """
    It computes the sequence order coupling numbers from 1 to max lag
    for a given protein sequence based on the distance matrix

    Parameters
    ----------
    protein_sequence : str
        A protein sequence

    max_lag : int
        Maximum lag to compute the sequence order coupling numbers

    distance_matrix : dict
        A distance matrix

    Returns
    -------
    sequence_order_coupling_number : List[float]
        List of sequence order coupling numbers
    """
    tau = []
    for i in range(max_lag):
        pairs = pairwise(protein_sequence, i + 1)
        tau.append(sum(distance_matrix[(a, b)] for a, b in pairs))
    return tau


class SequenceOrderCouplingNumbersDescriptor(Descriptor):
    """
    It calculates the Sequence Order Coupling Numbers of a sequence.
    """

    def __call__(self, sequence: str, max_lag: int = 45, **kwargs):
        """
        It calculates the Sequence Order Coupling Numbers of a sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the Sequence Order Coupling Numbers for.

        max_lag : int
            The maximum lag to consider. It should be inferior to the length of the sequence.

        kwargs

        Returns
        -------
        list of float
            the Sequence Order Coupling Numbers of a sequence.
        """
        if len(sequence) < max_lag:
            raise ValueError("The sequence length should be superior to the max_lag.")

        tau_sw = sequence_order_coupling_number(sequence, max_lag, distance_matrix=DISTANCE_1_POW_2)
        tau_grant = sequence_order_coupling_number(sequence, max_lag, distance_matrix=DISTANCE_2_POW_2)
        return tau_sw + tau_grant

    def get_features_out(self, max_lag: int = 45, **kwargs):
        return [f'tau_sw_{i}' for i in range(1, max_lag + 1)] + [f'tau_grant_{i}' for i in range(1, max_lag + 1)]


def quasi_sequence_order_1(protein_sequence: str,
                           max_lag: int = 30,
                           weight: float = 0.1,
                           distance_matrix: dict = DISTANCE_1_POW_2) -> List[float]:
    """
    It computes the quasi sequence order coupling numbers from 1 to max lag
    for a given protein sequence based on the distance matrix

    Parameters
    ----------
    protein_sequence : str
        A protein sequence

    max_lag : int
        Maximum lag to compute the quasi sequence order coupling numbers

    weight : float
        Weight of the amino acid composition descriptor

    distance_matrix : dict
        A distance matrix

    Returns
    -------
    quasi_sequence_order : List[float]
        List of quasi sequence order coupling numbers
    """
    sequence_order_factor = sum(sequence_order_coupling_number(protein_sequence, max_lag, distance_matrix))
    sequence_order_factor = 1 + (weight * sequence_order_factor)

    aac_descriptor = AminoAcidCompositionDescriptor()
    aac_cnt = aac_descriptor(protein_sequence)
    aac_keys = aac_descriptor.get_features_out()
    aac = dict(zip(aac_keys, aac_cnt))

    tau = []
    for i, aa in enumerate(AA_ALPHABET):
        tau.append(round(aac[aa] / sequence_order_factor, 6))
    return tau


def quasi_sequence_order_2(protein_sequence: str,
                           max_lag: int = 30,
                           weight: float = 0.1,
                           distance_matrix: dict = DISTANCE_1_POW_2) -> List[float]:
    """
    It computes the quasi sequence order coupling numbers from 1 to max lag
    for a given protein sequence based on the distance matrix

    Parameters
    ----------
    protein_sequence : str
        A protein sequence

    max_lag : int
        Maximum lag to compute the quasi sequence order coupling numbers

    weight : float
        Weight of the amino acid composition descriptor

    distance_matrix : dict
        A distance matrix

    Returns
    -------
    quasi_sequence_order : List[float]
        List of quasi sequence order coupling numbers
    """
    sequence_order_factors = sequence_order_coupling_number(protein_sequence, max_lag, distance_matrix)
    factor = 1 + (weight * sum(sequence_order_factors))

    tau = []
    for i in range(max_lag):
        tau.append(round(weight * sequence_order_factors[1] / factor, 6))
    return tau


class QuasiSequenceOrderDescriptor(Descriptor):
    """
    It calculates the Quasi-Sequence Order of a sequence.
    """

    def __call__(self, sequence: str, max_lag: int = 30, weight: float = 0.1, **kwargs):
        """
        It calculates the Quasi-Sequence Order of a sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the Quasi-Sequence Order for.

        max_lag : int
            The maximum lag to consider. It should be inferior to the length of the sequence.

        weight : float
            The weight to apply to the quasi-sequence order.

        kwargs

        Returns
        -------
        list of float
            the Quasi-Sequence Order of a sequence.
        """
        if len(sequence) < max_lag:
            raise ValueError("The sequence length should be superior to the max_lag.")

        tau_qso_1_sw = quasi_sequence_order_1(sequence, max_lag, distance_matrix=DISTANCE_1_POW_2)
        tau_qso_1_grant = quasi_sequence_order_1(sequence, max_lag, distance_matrix=DISTANCE_2_POW_2)
        tau_qso_2_sw = quasi_sequence_order_2(sequence, max_lag, distance_matrix=DISTANCE_1_POW_2)
        tau_qso_2_grant = quasi_sequence_order_2(sequence, max_lag, distance_matrix=DISTANCE_2_POW_2)
        return tau_qso_1_sw + tau_qso_2_sw + tau_qso_1_grant + tau_qso_2_grant

    def get_features_out(self, max_lag: int = 30, **kwargs):
        tau_qso_1_sw = [f"tau_qso_1_sw_{aa}" for aa in range(len(AA_ALPHABET))]
        tau_qso_1_grant = [f"tau_qso_1_grant_{aa}" for aa in range(len(AA_ALPHABET))]
        tau_qso_2_sw = [f"tau_qso_2_sw_{i}" for i in range(max_lag)]
        tau_qso_2_grant = [f"tau_qso_2_grant_{i}" for i in range(max_lag)]
        return tau_qso_1_sw + tau_qso_2_sw + tau_qso_1_grant + tau_qso_2_grant




import numpy as np

from plants_sm.data_structures.dataset import Dataset
from plants_sm.featurization.featurizer import FeaturesGenerator
from plants_sm.featurization.proteins.encodings._constants import ZS


class ZScaleEncoder(FeaturesGenerator):
    """
    It encodes protein sequences using the Z scale.
    This encoding requires the preprocessing of the protein
    sequences to the 20 standard amino acids.

    This method encodes which amino acid of the sequence into a Z-scales. Each Z scale represent an amino-acid property:
        Z1: Lipophilicity
        Z2: Steric properties (Steric bulk/Polarizability)
        Z3: Electronic properties (Polarity / Charge)
        Z4 and Z5: They relate electronegativity, heat of formation, electrophilicity and hardness.
    """

    output_shape_dimension = 3
    name = "z_scale"

    def set_features_names(self):
        self.features_names = ['Z1', 'Z2', 'Z3', 'Z4', 'Z5']

    def _fit(self, dataset: Dataset, instance_type: str) -> 'ZScaleEncoder':
        return self

    def _fit_batch(self, dataset: Dataset, instance_type: str) -> 'ZScaleEncoder':
        return self

    def _featurize(self, sequence: str) -> np.ndarray:
        """
        It encodes a protein sequence with the ZScale encoding.

        Parameters
        ----------
        sequence: str
            The protein sequence to be encoded.

        Returns
        -------
        encoded_sequence: np.ndarray
            The encoded sequence.
        """
        return np.array([np.array(ZS[aa]) for aa in sequence], dtype=np.float32)



import os
import warnings
from typing import Iterable

import numpy as np

from plants_sm.data_structures.dataset import Dataset
from plants_sm.io.pickle import write_pickle, is_pickable
from plants_sm.models.constants import QUANTILE, REGRESSION, BINARY, FileConstants


def _convert_proba_to_unified_form(problem_type, y_pred_proba: np.ndarray) -> np.ndarray:
    """
    Ensures that y_pred_proba is in a consistent form across all models. For binary classification,
    converts y_pred_proba to a 1 dimensional array of prediction probabilities of the positive class. For
    multiclass and softclass classification, keeps y_pred_proba as a 2 dimensional array of prediction
    probabilities for each class. For regression, converts y_pred_proba to a 1 dimensional array of predictions.

    Parameters
    ----------
    y_pred_proba: np.ndarray
        Array of prediction probabilities

    Returns
    -------
    np.ndarray
        Array of prediction probabilities in a consistent form across all models
    """
    if problem_type == REGRESSION:
        if len(y_pred_proba.shape) == 1:
            return y_pred_proba
        else:
            return y_pred_proba[:, 1]
    elif problem_type == BINARY:
        if len(y_pred_proba.shape) == 1:
            return y_pred_proba
        elif y_pred_proba.shape[1] > 1 and y_pred_proba.shape[1] == 1:
            return y_pred_proba[:, 1]
        else:
            return y_pred_proba
    elif y_pred_proba.shape[1] > 2:  # Multiclass, Softclass
        return y_pred_proba
    else:  # Unknown problem type
        raise AssertionError(f'Unknown y_pred_proba format for `problem_type="{problem_type}"`.')

def _get_pred_from_proba(problem_type, y_pred_proba: np.ndarray) -> np.ndarray:
    """
    Get the prediction from the probability

    Parameters
    ----------
    y_pred_proba: list
        List of probabilities
    Returns
    -------
    np.ndarray
        Array of predictions
    """
    if problem_type == BINARY:
        if y_pred_proba.shape[1] == 1:
            y_pred = np.array([1 if pred >= 0.5 else 0 for pred in y_pred_proba])
        else:
            y_pred = multi_label_binarize(y_pred_proba)
    elif problem_type == REGRESSION:
        y_pred = y_pred_proba
    elif problem_type == QUANTILE:
        y_pred = y_pred_proba
    else:
        y_pred = []
        if not len(y_pred_proba) == 0:
            y_pred = np.argmax(y_pred_proba, axis=1)
            return y_pred

    y_pred = array_reshape(y_pred)
    return y_pred

def write_model_parameters_to_pickle(model_parameters: dict, path: str) -> None:
    """
    Write the model parameters to a pickle file.

    Parameters
    ----------
    model_parameters: dict
        Dictionary of model parameters
    path: str
        Path to save the model
    """
    parameters = {}
    for key, value in model_parameters.items():
        if is_pickable(value):
            parameters[key] = value
        else:
            warning_str = f'Could not save {key} to save file. Skipping attribute {key}.'
            warnings.warn(warning_str)
    write_pickle(os.path.join(path, FileConstants.MODEL_PARAMETERS_PKL.value), parameters)


def array_from_tensor(tensor) -> np.ndarray:
    """
    Converts a tensor to a numpy array.

    Parameters
    ----------
    tensor: Tensor
        The tensor to convert.

    Returns
    -------
    np.ndarray
        The converted array.
    """
    tensor = tensor.cpu().detach().numpy()
    return tensor


def array_reshape(array: np.ndarray) -> np.ndarray:
    """
    Reshapes a numpy array.

    Parameters
    ----------
    array: np.ndarray
        The array to reshape.

    Returns
    -------
    np.ndarray
        The reshaped array.
    """
    # Reshape array if necessary if it is a 1D array or if it is a 2D array with only 1 column
    if len(array.shape) == 1 or (len(array.shape) == 2 and array.shape[1] == 1):
        array = array.reshape(-1, 1)
    return array


def multi_label_binarize(y_pred_proba, threshold=0.5) -> np.ndarray:
    """
    Binarize the predicted probabilities for multi-label classification.

    Parameters
    ----------
    y_pred_proba: np.ndarray
        Predicted probabilities with shape (n_samples, n_classes).
    threshold: float
        Threshold for binarization.

    Returns:
        np.ndarray: Binary predictions with shape (n_samples, n_classes).
    """
    n_samples, n_classes = y_pred_proba.shape
    y_pred = np.zeros((n_samples, n_classes))
    for i in range(n_classes):
        y_pred[:, i] = (y_pred_proba[:, i] >= threshold).astype(int)
    return y_pred


def _batch_generator(dataset: Dataset) -> Iterable:
    """
    Generate batches of data.

    Parameters
    ----------
    dataset: Dataset
        Dataset to be used for training.

    Returns
    -------
    Union[list, dict]
        Batches of data.
    """
    while dataset.next_batch():
        if dataset.y is None:
            yield dataset.X
        else:
            yield dataset.X, dataset.y



import io
import logging
import os
from logging.handlers import TimedRotatingFileHandler
import pickle
from typing import Callable, Union, Tuple, List, Dict

import numpy as np
import pandas as pd
from numpy import ndarray
from torch import nn, Tensor
from torch.nn.modules.loss import _Loss
from torch.optim import Adam, Optimizer
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torch.utils.data import TensorDataset, DataLoader
from torch.utils.tensorboard import SummaryWriter

from plants_sm.data_structures.dataset import Dataset
from plants_sm.io.pickle import read_pickle, write_pickle
from plants_sm.models._utils import _convert_proba_to_unified_form, \
    write_model_parameters_to_pickle, array_from_tensor, array_reshape, multi_label_binarize
from plants_sm.models.constants import REGRESSION, QUANTILE, BINARY, FileConstants
from plants_sm.models.model import Model
import torch


class PyTorchModel(Model):

    def __init__(self, model: nn.Module, loss_function: _Loss, validation_loss_function: _Loss = None, model_name=None,
                 optimizer: Optimizer = None,
                 scheduler: ReduceLROnPlateau = None, epochs: int = 32, batch_size: int = 32,
                 patience: int = 4, validation_metric: Callable = None,
                 problem_type: str = BINARY,
                 device: Union[str, torch.device, List[str]] = torch.device(
                     'cuda' if torch.cuda.is_available() else 'cpu'),
                 trigger_times: int = 0,
                 last_loss: int = None,
                 progress: int = 100,
                 logger_path: str = None, tensorboard_file_path: str = None,
                 l2_regularization_lambda: float = None, l1_regularization_lambda: float = None,
                 checkpoints_path: str = None,
                 early_stopping_method="loss",
                 objective="min"):

        """
        Constructor for PyTorchModel

        Parameters
        ----------
        model: nn.Module
            PyTorch model
        loss_function: _Loss
            PyTorch loss function
        validation_loss_function: _Loss
            PyTorch loss function for validation
        optimizer: Optimizer
            PyTorch optimizer
        scheduler: ReduceLROnPlateau
            PyTorch scheduler
        epochs: int
            Number of epochs
        batch_size: int
            Batch size
        patience: int
            Number of epochs to wait before early stopping
        validation_metric: Callable
            Sklearn metric function to use for validation
        problem_type: str
            Type of problem
        device: Union[str, torch.device]
            Device to use for training
        trigger_times: int
            Number of times the model has been triggered
        last_loss: int
            Last loss value
        progress: int
            Number of batches to wait before logging progress
        logger_path: str
            Path to save the logger
        tensorboard_file_path: str
            Path to save the tensorboard file
        l2_regularization_lambda: float
            L2 regularization lambda
        l1_regularization_lambda: float
            L1 regularization lambda
        checkpoints_path: str
            Path to save the checkpoints
        early_stopping_method: str
            Early stopping metric: Options: loss, metric
        objective: str
            Objective of the early stopping metric: Options: min, max
        """

        super().__init__()

        if model_name:
            self.model_name = model_name
        else:
            self.model_name = model.__class__.__name__

        if isinstance(device, list):
            self.model = model
            self.device = device

        else:
            self.device = device
            self.model = model.to(self.device)

        self.logger_path = logger_path
        if tensorboard_file_path is None:
            self.tensorboard_file_path = "./runs"
        else:
            self.tensorboard_file_path = tensorboard_file_path

        self.loss_function = loss_function

        if validation_loss_function is None:
            self.validation_loss_function = loss_function
        else:
            self.validation_loss_function = validation_loss_function

        self.optimizer = optimizer
        self.progress = progress
        self.scheduler = scheduler
        self.epochs = epochs
        self.batch_size = batch_size
        self.patience = patience
        self.validation_metric = validation_metric
        self.problem_type = problem_type
        self.trigger_times = trigger_times
        self.last_loss = last_loss

        loss_dataframe = pd.DataFrame(columns=["epoch", "train_loss", "valid_loss"])
        loss_dataframe.set_index("epoch", inplace=True)

        metric_dataframe = pd.DataFrame(columns=["epoch", "train_metric_result", "valid_metric_result"])
        metric_dataframe.set_index("epoch", inplace=True)

        self._history = {'loss': loss_dataframe,
                         'metric_results': metric_dataframe}

        if not self.optimizer:
            self.optimizer = Adam(self.model.parameters())

        self.losses = {}
        self.metrics = {}

        self.l2_regularization_lambda = l2_regularization_lambda
        self.l1_regularization_lambda = l1_regularization_lambda

        if checkpoints_path is None:
            checkpoints_path = "./checkpoints"
        self.checkpoints_path = checkpoints_path

        if objective == "max":
            self.best_metric_result = -np.inf
        elif objective == "min":
            self.best_metric_result = np.inf
        else:
            raise ValueError("Objective must be either min or max")

        self.early_stopping_method = early_stopping_method
        self.objective = objective

    @property
    def history(self) -> dict:
        """
        Get the history of the model

        Returns
        -------
        dict
            History of the model
        """
        return self._history

    @staticmethod
    def _read_pytorch_model(path: str) -> nn.Module:
        """
        Read the model from the specified path.

        Parameters
        ----------
        path: str
            Path to read the model from

        Returns
        -------
        torch.nn.Module
        """
        weights_path = os.path.join(path, FileConstants.PYTORCH_MODEL_WEIGHTS.value)
        try:
            model = read_pickle(os.path.join(path, FileConstants.PYTORCH_MODEL_PKL.value))
        except RuntimeError:
            class CPU_Unpickler(pickle.Unpickler):
                def find_class(self, module, name):
                    if module == 'torch.storage' and name == '_load_from_bytes':
                        return lambda b: torch.load(io.BytesIO(b), map_location='cpu')
                    else:
                        return super().find_class(module, name)

            with open(os.path.join(path, FileConstants.PYTORCH_MODEL_PKL.value), "rb") as f:
                model = CPU_Unpickler(f).load()
        model.load_state_dict(torch.load(weights_path, map_location=torch.device('cpu')))
        model.eval()
        return model

    @staticmethod
    def _save_pytorch_model(model: nn.Module, path: str) -> None:
        """
        Save the model to the specified path.

        Parameters
        ----------
        model: torch.nn.Module
            Model to be saved
        path: str
            Path to save the model

        Returns
        -------

        """
        weights_path = os.path.join(path, FileConstants.PYTORCH_MODEL_WEIGHTS.value)
        torch.save(model.state_dict(), weights_path)
        write_pickle(os.path.join(path, FileConstants.PYTORCH_MODEL_PKL.value), model)

    def _save(self, path: str):
        """
        Save the model to a file
        This method is called by the save method and needs to have all the parameters one wants to save in the model.

        Parameters
        ----------
        path: str
            Path to save the model

        Returns
        -------
        """

        self._save_pytorch_model(self.model, path)

        model_parameters = {
            'loss_function': self.loss_function,
            'optimizer': self.optimizer,
            'scheduler': self.scheduler,
            'epochs': self.epochs,
            'batch_size': self.batch_size,
            'patience': self.patience,
            'validation_metric': self.validation_metric,
            'problem_type': self.problem_type,
            'device': self.device,
            'trigger_times': self.trigger_times,
            'last_loss': self.last_loss,
            'progress': self.progress
        }

        write_model_parameters_to_pickle(model_parameters, path)

    @classmethod
    def _load(cls, path: str) -> 'PyTorchModel':
        """
        Load the model from a file

        Parameters
        ----------
        path: str
            Path to load the model

        """
        model = cls._read_pytorch_model(path)
        model_parameters = read_pickle(os.path.join(path, FileConstants.MODEL_PARAMETERS_PKL.value))
        return cls(model, **model_parameters)

    def _preprocess_data(self, dataset: Dataset, shuffle: bool = True) -> DataLoader:
        """
        Preprocess the data for training

        Parameters
        ----------
        dataset: Dataset
            Dataset to preprocess
        shuffle: bool
            Whether to shuffle the data

        Returns
        -------
        DataLoader
            Preprocessed data
        """

        tensors = []
        if isinstance(dataset.X, Dict):
            for instance in dataset.X.keys():
                tensor = torch.tensor(dataset.X[instance], dtype=torch.float)
                tensors.append(tensor)
        else:
            tensor = torch.tensor(dataset.X, dtype=torch.float)
            tensors.append(tensor)

        try:
            if dataset.y is not None:
                tensors.append(torch.tensor(dataset.y, dtype=torch.float))
        except ValueError:
            pass
        dataset = TensorDataset(
            *tensors
        )

        data_loader = DataLoader(
            dataset,
            shuffle=shuffle,
            batch_size=self.batch_size
        )
        return data_loader

    def _validate_batches(self, validation_set: DataLoader, loss_total: float,
                          predictions: ndarray,
                          actuals: ndarray) -> Tuple[float, ndarray, ndarray]:
        """
        Validate the model on the validation set

        Parameters
        ----------
        validation_set: DataLoader
            Validation set
        loss_total: float
            Total loss
        predictions: ndarray
            Predictions
        actuals: ndarray
            Actual values

        Returns
        -------
        Tuple[float, ndarray, ndarray]
            Total loss, predictions and actual values
        """

        for i, inputs_targets in enumerate(validation_set):
            inputs, targets = inputs_targets[:-1], inputs_targets[-1]
            for j, inputs_elem in enumerate(inputs):
                inputs[j] = inputs_elem.to(self.device)

            targets = targets.to(self.device)
            output = self.model(inputs)

            y_hat = array_from_tensor(output)
            actual = array_from_tensor(targets)

            predictions = np.concatenate((predictions, y_hat))
            actuals = np.concatenate((actuals, actual))

            loss = self.validation_loss_function(output, targets)
            loss_total += loss.item()

        return loss_total, predictions, actuals

    def _validate(self, validation_set: Dataset) -> Tuple[float, float]:
        """
        Validate the model

        Parameters
        ----------
        validation_set: DataLoader
            Validation set

        Returns
        -------
        Tuple[float, float]
            Validation loss and validation metric result
        """
        self.model.eval()
        loss_total = 0
        second_shape = validation_set.y.shape[1]
        predictions, actuals = np.empty(shape=(0, second_shape)), np.empty(shape=(0, second_shape))

        predictions = array_reshape(predictions)
        actuals = array_reshape(actuals)

        with torch.no_grad():

            if validation_set.batch_size is None:
                validation_set_preprocessed = self._preprocess_data(validation_set, shuffle=False)
                len_valid_dataset = len(validation_set_preprocessed)
                loss_total, predictions, actuals = \
                    self._validate_batches(validation_set_preprocessed, loss_total,
                                           predictions, actuals)

            else:
                len_valid_dataset = 0
                while validation_set.next_batch():
                    validation_set_preprocessed = self._preprocess_data(validation_set, shuffle=False)
                    len_valid_dataset += len(validation_set_preprocessed)
                    loss_total, predictions, actuals = \
                        self._validate_batches(validation_set_preprocessed, loss_total,
                                               predictions, actuals)

        validation_metric_result = None
        if self.validation_metric:
            predictions = self.get_pred_from_proba(np.array(predictions))
            validation_metric_result = self.validation_metric(actuals, predictions)

        return loss_total / len_valid_dataset, validation_metric_result

    @staticmethod
    def _train_ddp(inputs_targets: Tensor, model, loss_function, optimizer,
                   l2_regularization_lambda, l1_regularization_lambda, device):

        inputs, targets = inputs_targets[:-1], inputs_targets[-1]
        for j, inputs_elem in enumerate(inputs):
            inputs[j] = inputs_elem.to(device[0])

        targets = targets.to(device[0])

        optimizer.zero_grad()

        output = model(inputs)
        output = output.to(device[0])

        # Forward and backward propagation
        loss = loss_function(output, targets)

        if l2_regularization_lambda:
            l2_lambda = l2_regularization_lambda
            l2_norm = sum(p.pow(2.0).sum()
                          for p in model.parameters())

            loss = loss + l2_lambda * l2_norm

        if l1_regularization_lambda:
            l1_lambda = l1_regularization_lambda
            l1_norm = sum(p.abs().sum()
                          for p in model.parameters())

            loss = loss + l1_lambda * l1_norm

        # model.to(device[0])
        loss.backward()
        optimizer.step()

        actual = array_from_tensor(targets)
        actual = array_reshape(actual)

        yhat = array_from_tensor(output)
        yhat = array_reshape(yhat)

        return actual, yhat, loss, model, optimizer

    def _train(self, inputs_targets: Tensor) -> Tuple[np.ndarray, np.ndarray, Tensor]:
        """
        Train the model

        Parameters
        ----------
        inputs_targets: Tensor
            Inputs and targets

        Returns
        -------
        Tuple[List[float], List[float], Tensor]
            Loss, predictions, targets
        """

        inputs, targets = inputs_targets[:-1], inputs_targets[-1]
        for j, inputs_elem in enumerate(inputs):
            inputs[j] = inputs_elem.to(self.device)

        targets = targets.to(self.device)

        self.optimizer.zero_grad()

        output = self.model(inputs)

        # Forward and backward propagation
        loss = self.loss_function(output, targets)

        if self.l2_regularization_lambda:
            l2_lambda = self.l2_regularization_lambda
            l2_norm = sum(p.pow(2.0).sum()
                          for p in self.model.parameters())

            loss = loss + l2_lambda * l2_norm

        if self.l1_regularization_lambda:
            l1_lambda = self.l1_regularization_lambda
            l1_norm = sum(p.abs().sum()
                          for p in self.model.parameters())

            loss = loss + l1_lambda * l1_norm

        loss.backward()
        self.optimizer.step()

        actual = array_from_tensor(targets)
        actual = array_reshape(actual)

        yhat = array_from_tensor(output)
        yhat = array_reshape(yhat)

        return actual, yhat, loss

    def _register_history(self, loss: float, epoch: int, metric_result: float = None, train: bool = True) -> None:
        """
        Register the history of the model

        Parameters
        ----------
        loss: float
            Loss
        epoch: int
            Epoch
        metric_result: float
            Metric result
        train: bool
            Whether it is training or validation
        """
        dataset_type = "train" if train else "valid"

        self.writer.add_scalar(f"Loss/{dataset_type}", loss, epoch)
        self._history["loss"].at[epoch - 1, f"{dataset_type}_loss"] = loss

        if metric_result is not None:
            self.writer.add_scalar(f"Metric/{dataset_type}", metric_result, epoch)
            self._history["metric_results"].at[epoch - 1, f"{dataset_type}_metric_result"] = metric_result

    def _early_stopping(self, validation_dataset: Dataset, epoch: int) -> Union[nn.Module, None]:
        """
        Early stopping

        Parameters
        ----------
        validation_dataset: DataLoader
            Validation dataset
        epoch: int
            Epoch

        Returns
        -------
        Union[nn.Module, None]
            Model or None
        """

        # Early stopping
        current_loss, validation_metric_result = self._validate(validation_dataset)

        self.losses[epoch] = current_loss
        self.metrics[epoch] = validation_metric_result

        if validation_metric_result is not None:
            self.logger.info(
                f'Validation metric: {validation_metric_result:.8}')

        self.logger.info(f'Validation loss: {current_loss:.8}')

        if self.early_stopping_method == "loss":
            early_stop = current_loss >= self.last_loss
        elif self.early_stopping_method == "metric":
            early_stop = validation_metric_result <= self.best_metric_result
        else:
            raise ValueError(f"early_stopping_method must be 'loss' or 'metric', got {self.early_stopping_method}")

        if early_stop:
            self.trigger_times += 1

            if self.trigger_times >= self.patience:
                self.logger.info(f'Early stopping at epoch {epoch}')
                self._register_history(current_loss, epoch, validation_metric_result, train=False)

                # choose the best model based on the validation loss
                # get the key of a dictionary whose value is the minimum
                best_epoch = self._get_best_epoch()
                self.logger.info(f'Best epoch: {best_epoch}')
                self.model.load_state_dict(
                    torch.load(f"{self.checkpoints_path}/{self.model_name}/epoch_{best_epoch}/model.pt"))

                return self.model

        else:
            self.trigger_times = 0

        self._register_history(current_loss, epoch, validation_metric_result, train=False)

        self.last_loss = current_loss

    def _calculate_metric_result(self, actuals: np.ndarray, predictions: np.ndarray) -> float:
        validation_metric_result = None
        if self.validation_metric:
            predictions = self.get_pred_from_proba(np.array(predictions))
            validation_metric_result = self.validation_metric(actuals, predictions)

        return validation_metric_result

    def _train_epoch(self, train_dataset: Dataset, epoch: int,
                     validation_dataset: Dataset = None) -> Union[nn.Module, None]:
        """
        Train the model for one epoch

        Parameters
        ----------
        train_dataset: Dataset
            Training dataset
        epoch: int
            Epoch
        validation_dataset: Dataset
            Validation dataset

        Returns
        -------
        Union[nn.Module, None]
            Model
        """
        self.model.train()
        loss_total = 0

        second_shape = train_dataset.y.shape[1]
        predictions, actuals = np.empty(shape=(0, second_shape)), np.empty(shape=(0, second_shape))

        predictions = array_reshape(predictions)
        actuals = array_reshape(actuals)

        train_dataset_preprocessed = self._preprocess_data(train_dataset, shuffle=False)

        len_train_dataset = len(train_dataset_preprocessed)

        for i, inputs_targets in enumerate(train_dataset_preprocessed):
            actual, yhat, loss = self._train(inputs_targets)

            predictions = np.concatenate((predictions, yhat))
            actuals = np.concatenate((actuals, actual))
            loss_total += loss.item()
            # Show progress
            if i % self.progress == 0 or i == len_train_dataset - 1:
                loss_partial = loss_total / (i + 1)
                self.logger.info(f'[{epoch}/{self.epochs}, {i}/{len_train_dataset}] loss: {loss_partial:.8}')

        loss = loss_total / len_train_dataset

        if self.validation_metric:
            predictions = self.predict(train_dataset)
            validation_metric_result = self._calculate_metric_result(actuals, predictions)

            if validation_metric_result is not None:
                self.logger.info(f'[{epoch}/{self.epochs}] metric result: {validation_metric_result:.8}')
        else:
            validation_metric_result = None

        self.logger.info(
            f'[{epoch}/{self.epochs}] Training loss: {loss:.8}')
        self._register_history(loss, epoch, validation_metric_result)

        if validation_dataset:
            assert validation_dataset != train_dataset, "Validation dataset should not be the same as training dataset"
            return self._early_stopping(validation_dataset, epoch)
        else:
            return None

    def _train_epoch_batch(self, train_dataset: Dataset, epoch: int, batch_num: int, loss_total_whole_dataset: float) \
            -> Tuple[float, np.ndarray, np.ndarray]:
        """
        Train the model for one epoch

        Parameters
        ----------
        train_dataset: Dataset
            Training dataset
        epoch: int
            Epoch
        batch_num: int
            Batch number

        Returns
        -------
        Union[nn.Module, None]
            Model
        """
        self.model.train()
        loss_total = 0

        second_shape = train_dataset.y.shape[1]
        predictions, actuals = np.empty(shape=(0, second_shape)), np.empty(shape=(0, second_shape))

        predictions = array_reshape(predictions)
        actuals = array_reshape(actuals)

        train_dataset_preprocessed = self._preprocess_data(train_dataset, shuffle=False)

        len_train_dataset = len(train_dataset_preprocessed)

        for i, inputs_targets in enumerate(train_dataset_preprocessed):
            actual, yhat, loss = self._train(inputs_targets)

            predictions = np.concatenate((predictions, yhat))
            actuals = np.concatenate((actuals, actual))
            loss_total += loss.item()
            # Show progress

        loss = loss_total / len_train_dataset
        loss_partial = (loss_total_whole_dataset + loss) / batch_num
        self.logger.info(f'[{epoch}/{self.epochs}, batch '
                         f'{batch_num}] loss: {loss_partial:.8}')

        return loss, predictions, actuals

    def _fit_data(self, train_dataset: Dataset, validation_dataset: Dataset = None) -> nn.Module:
        """
        Fit the model to the data

        Parameters
        ----------
        train_dataset: Dataset
            Training dataset
        validation_dataset: Dataset
            Validation dataset

        Returns
        -------
        nn.Module
            Trained model
        """

        self.writer = SummaryWriter(log_dir=self.tensorboard_file_path, comment=self.model_name,
                                    filename_suffix=self.model_name)

        formatter = logging.Formatter('%(asctime)s %(name)s %(levelname)s %(message)s')
        if self.logger_path:
            handler = TimedRotatingFileHandler(self.logger_path, when='midnight', backupCount=30)
        else:
            handler = TimedRotatingFileHandler(f'./{self.model_name}.log', when='midnight', backupCount=20)
        handler.setFormatter(formatter)

        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(handler)
        self.logger.setLevel(logging.DEBUG)

        self.last_loss = 100
        self.trigger_times = 0

        self.logger.info("starting to fit the data...")

        if train_dataset.batch_size is None:

            for epoch in range(1, self.epochs + 1):
                self._train_epoch(train_dataset, epoch, validation_dataset)

                self._write_model_check_points(epoch)

                if self.scheduler is not None:
                    self.scheduler.step(self.last_loss)
        else:
            second_shape = train_dataset.y.shape[1]

            for epoch in range(1, self.epochs + 1):
                batch_i = 0
                loss_total = 0

                predictions, actuals = np.empty(shape=(0, second_shape)), np.empty(shape=(0, second_shape))
                actuals_final = array_reshape(actuals)

                while train_dataset.next_batch():
                    batch_i += 1
                    loss, predictions, actuals = self._train_epoch_batch(train_dataset, epoch, batch_i, loss_total)
                    actuals_final = np.concatenate((actuals_final, actuals))

                    loss_total += loss

                loss = loss_total / batch_i

                predictions_final = self.predict(train_dataset)

                validation_metric_result = self._calculate_metric_result(actuals_final, predictions_final)
                if validation_metric_result is not None:
                    self.logger.info(f'[{epoch}/{self.epochs}] metric result: {validation_metric_result:.8}')

                self.logger.info(
                    f'[{epoch}/{self.epochs}] Training loss: {loss:.8}')
                self._register_history(loss, epoch, validation_metric_result)

                self._write_model_check_points(epoch)

                if self.scheduler is not None:
                    self.scheduler.step(self.last_loss)
                if validation_dataset:
                    assert validation_dataset != train_dataset, "Validation dataset should not be the same as " \
                                                                "training dataset"
                    model = self._early_stopping(validation_dataset, epoch)

                    if model:
                        self.writer.flush()
                        return model

        if validation_dataset:
            best_epoch = self._get_best_epoch()
            self.logger.info(f'Best epoch: {best_epoch}')
            self.model.load_state_dict(
                torch.load(f"{self.checkpoints_path}/{self.model_name}/epoch_{best_epoch}/model.pt"))

        self.writer.flush()
        return self.model

    def _get_best_epoch(self) -> int:
        if self.early_stopping_method == "loss":
            if self.objective == "min":
                best_epoch = min(self.losses, key=self.losses.get)
            elif self.objective == "max":
                best_epoch = max(self.losses, key=self.losses.get)
            else:
                raise ValueError("Objective not recognized. It should be either 'min' or 'max'")

            self.logger.info(f'Best loss: {self.losses[best_epoch]}')
        elif self.early_stopping_method == "metric":
            if self.objective == "min":
                best_epoch = min(self.metrics, key=self.metrics.get)
            elif self.objective == "max":
                best_epoch = max(self.metrics, key=self.metrics.get)
            else:
                raise ValueError("Objective not recognized. It should be either 'min' or 'max'")
            self.logger.info(f'Best metric value: {self.metrics[best_epoch]}')
        else:
            raise ValueError("Early stopping method not recognized")

        return best_epoch

    def _write_model_check_points(self, epoch: int) -> None:
        """
        Write the model checkpoints for tensorboard.

        Parameters
        ----------
        epoch: int
            Epoch
        """

        os.makedirs(self.checkpoints_path, exist_ok=True)
        os.makedirs(f"{self.checkpoints_path}/{self.model_name}/epoch_{epoch}", exist_ok=True)
        torch.save(self.model.state_dict(), f"{self.checkpoints_path}/{self.model_name}/epoch_{epoch}"
                                            f"/model.pt")

    def get_pred_from_proba(self, y_pred_proba: np.ndarray) -> np.ndarray:
        """
        Get the prediction from the probability

        Parameters
        ----------
        y_pred_proba: list
            List of probabilities
        Returns
        -------
        np.ndarray
            Array of predictions
        """
        if self.problem_type == BINARY:
            if y_pred_proba.shape[1] == 1:
                y_pred = np.array([1 if pred >= 0.5 else 0 for pred in y_pred_proba])
            else:
                y_pred = multi_label_binarize(y_pred_proba)
        elif self.problem_type == REGRESSION:
            y_pred = y_pred_proba
        elif self.problem_type == QUANTILE:
            y_pred = y_pred_proba
        else:
            y_pred = []
            if not len(y_pred_proba) == 0:
                y_pred = np.argmax(y_pred_proba, axis=1)
                return y_pred

        y_pred = array_reshape(y_pred)
        return y_pred

    def _predict_proba_batch(self, dataset: Dataset, predictions: np.ndarray = None) -> np.ndarray:
        """
        Predicts the probability of each class for each sample in the dataset.

        Parameters
        ----------
        dataset: DataLoader
            Dataset to predict on
        predictions: np.ndarray
            Array of predictions

        Returns
        -------

        """
        dataset_preprocessed = self._preprocess_data(dataset, shuffle=False)
        for i, inputs_targets in enumerate(dataset_preprocessed):
            for j, inputs_elem in enumerate(inputs_targets):
                inputs_targets[j] = inputs_elem.to(self.device)

            yhat = self.model(inputs_targets)

            yhat = array_from_tensor(yhat)
            yhat = array_reshape(yhat)

            if i == 0 and predictions is None:
                predictions = yhat
            else:
                predictions = np.concatenate((predictions, yhat))

        return predictions

    def _predict_proba(self, dataset: Dataset) -> np.ndarray:
        """
        Predicts the probability of each class for each sample in the dataset.

        Parameters
        ----------
        dataset: Dataset
            Dataset to predict on

        Returns
        -------
        np.ndarray
            Array of prediction probabilities
        """

        if self.problem_type in [REGRESSION, QUANTILE]:
            y_pred = self.model.predict(dataset)
            return y_pred

        self.model.eval()

        predictions = None

        # the "shuffle" argument always has to be False in predicting probabilities in an evaluation context
        with torch.no_grad():

            if dataset.batch_size is None:
                predictions = self._predict_proba_batch(dataset)
            else:
                while dataset.next_batch():
                    predictions = self._predict_proba_batch(dataset, predictions)

        return _convert_proba_to_unified_form(self.problem_type, np.array(predictions))

    def _predict(self, dataset: Dataset) -> np.ndarray:
        """
        Predicts the class for each sample in the dataset.

        Parameters
        ----------
        dataset: Dataset
            Dataset to predict on

        Returns
        -------
        np.ndarray
            Array of predictions
        """
        y_pred_proba = self._predict_proba(dataset)
        y_pred = self.get_pred_from_proba(y_pred_proba)
        return y_pred



import torch
import torch.nn as nn


class HDMLF(nn.Module):
    def __init__(self, input_size, hidden_size, num_classes):
        super(HDMLF, self).__init__()

        # Input layer
        self.input_layer = nn.Linear(input_size, 1280)

        # Bidirectional GRU
        self.bi_gru = nn.GRU(1280, hidden_size, bidirectional=True, batch_first=True)

        # Attention layer
        self.attention = nn.MultiheadAttention(embed_dim=hidden_size * 2, num_heads=1)

        # Dense layer 1
        self.dense1 = nn.Linear(hidden_size * 2, 64)

        # Linear layer (output layer)
        self.linear = nn.Linear(64, num_classes)

    def forward(self, x):
        # Input layer
        x = self.input_layer(x)

        # Bidirectional GRU
        x, _ = self.bi_gru(x)

        # Attention layer
        x, _ = self.attention(x.permute(1, 0, 2), x.permute(1, 0, 2), x.permute(1, 0, 2))
        x = x.permute(1, 0, 2)

        # Global average pooling
        x = torch.mean(x, dim=1)

        # Dense layer 1
        x = self.dense1(x)

        # Linear layer (output layer)
        x = self.linear(x)

        return x


from typing import Tuple

from plants_sm.data_standardization.padding_enumerators import PaddingEnumerators
from plants_sm.data_structures.dataset import Dataset
from plants_sm.transformation._utils import transform_instances
from plants_sm.transformation.transformer import Transformer


class SMILESPadder(Transformer):
    """
    Padder for sequences of variable lengths.

    Attributes
    ----------
    pad_width: int
        width of the padded sequences
    padding: str
        type of padding to be applied
    """
    pad_width: int = None
    padding: str = "left"
    _pad_width_set: bool = False

    def __init__(self, pad_width: int = None, padding: str = "left"):
        super().__init__()
        if pad_width is not None:
            self._pad_width_set = True
        self.pad_width = pad_width
        self.padding = padding

    def _fit(self, dataset: Dataset, instance_type: str) -> 'SMILESPadder':
        """
        Method that fits the sequence padder to the dataset

        Parameters
        ----------
        dataset: Dataset
            dataset to fit the transformer where instances are the representation or object to be processed.

        Returns
        -------
        fitted sequence padder: SequencePadder
        """
        if not self.pad_width:
            # get the maximum length of the sequences
            lengths = [len(instance) for instance in dataset.get_instances(instance_type).values()]
            self.pad_width = max(lengths)

        if self.padding not in ["right", "left", "center"]:
            raise ValueError(f"Padding type not supported: {self.padding}")

        return self

    def _fit_batch(self, dataset: Dataset, instance_type: str) -> 'SMILESPadder':
        """
        Method that fits the sequence padder to the dataset

        Parameters
        ----------
        dataset: Dataset
            dataset to fit the transformer where instances are the representation or object to be processed.

        Returns
        -------
        fitted sequence padder: SequencePadder
        """
        if not self._pad_width_set:
            # get the maximum length of the sequences
            lengths = [len(instance) for instance in dataset.get_instances(instance_type).values()]
            max_length = max(lengths)
            if self.pad_width < max_length:
                self.pad_width = max_length

        if self.padding not in ["right", "left", "center"]:
            raise ValueError(f"Padding type not supported: {self.padding}")

        return self

    def _transform(self, dataset: Dataset, instance_type: str) -> Dataset:
        """
        Abstract method that has to be implemented by all feature generators

        Parameters
        ----------
        dataset: Dataset
            dataset to be transformed where instances are the representation or object to be processed.
        instance_type: str
            type of instances to be transformed

        Returns
        -------
        dataset with features: Dataset
            dataset object with features
        """
        return transform_instances(self.n_jobs, dataset, self._pad_sequence, instance_type,
                                   self.__class__.__name__)

    def _pad_sequence(self, instance: str, identifier: str) -> Tuple[str, str]:
        """
        Pad a sequence of variable length to a fixed length

        Parameters
        ----------
        instance: str
            sequence to be padded
        identifier: str
            identifier of the sequence

        Returns
        -------
        padded sequence: Dict[str, str]
            dictionary with the padded sequence
        """

        padded = None

        try:
            assert len(instance) <= self.pad_width, f"Sequence length is greater than pad width: " \
                                                    f"{len(instance)} > {self.pad_width}"

            if self.padding == "right":
                padded = instance.rjust(self.pad_width, str(PaddingEnumerators.COMPOUNDS.value))

            elif self.padding == "left":
                padded = instance.ljust(self.pad_width, str(PaddingEnumerators.COMPOUNDS.value))

            elif self.padding == "center":
                padded = instance.center(self.pad_width, str(PaddingEnumerators.COMPOUNDS.value))

        except AssertionError:

            padded = instance[:self.pad_width]

        return identifier, padded



from typing import List

import numpy as np

from plants_sm.data_structures.dataset import Dataset
from plants_sm.featurization.featurizer import FeaturesGenerator
from plants_sm.featurization.proteins.propythia.propythia_descriptors.presets import DESCRIPTORS_PRESETS


class PropythiaWrapper(FeaturesGenerator):

    preset: str = 'all'

    def set_features_names(self) -> List[str]:
        """
        The method features_names will return the names of the features.
        """
        self.features_names = []
        for descriptor in DESCRIPTORS_PRESETS[self.preset]:
            instantiated_descriptor = descriptor()
            self.features_names.extend(instantiated_descriptor.get_features_out())
        return self.features_names

    def _fit(self, dataset: Dataset, instance_type: str) -> 'PropythiaWrapper':
        """
        Abstract method that has to be implemented by all feature generators

        Parameters
        ----------
        dataset: Dataset
            dataset to fit the transformer where instances are the representation or object to be processed.
        """
        if self.preset not in DESCRIPTORS_PRESETS:
            raise ValueError(f'Preset {self.preset} is not available.')
        self.descriptors = []
        for descriptor in DESCRIPTORS_PRESETS[self.preset]:
            instantiated_descriptor = descriptor()
            self.descriptors.append(instantiated_descriptor)

        return self

    def _fit_batch(self, dataset: Dataset, instance_type: str) -> 'PropythiaWrapper':
        """
        Abstract method that has to be implemented by all feature generators

        Parameters
        ----------
        dataset: Dataset
            dataset to fit the transformer where instances are the representation or object to be processed.
        instance_type: str
            type of the instances to be featurized
        """
        return self._fit(dataset, instance_type)

    def _featurize(self, protein_sequence: str) -> np.ndarray:
        """
        The method _featurize will generate the desired features for a given protein sequence

        Parameters
        ----------
        protein_sequence: str
            protein sequence string

        Returns
        -------
        features_names: List[str]
            the names of the features

        features: np.ndarray
            the features
        """
        features_list = []
        for descriptor in self.descriptors:
            features = descriptor(protein_sequence)
            features_list.extend(features)

        return np.array(features_list, dtype=np.float32)



from joblib import Parallel, delayed
import torch
from tqdm import tqdm
from plants_sm.data_structures.dataset.dataset import Dataset
from plants_sm.data_structures.dataset.single_input_dataset import PLACEHOLDER_FIELD
from plants_sm.featurization.proteins.bio_embeddings.constants import ESM_FUNCTIONS
from plants_sm.models.lightning_model import InternalLightningModel
from torch.utils.data import DataLoader
from torch.utils.data import TensorDataset


class ESM(InternalLightningModel):

    def __init__(self, module, batch_size: int = 32, devices="cpu", **trainer_kwargs) -> None:

        self.model_name = module.model_name

        super().__init__(module, batch_size, devices, **trainer_kwargs)

    def _preprocess_data(self, dataset: Dataset, shuffle: bool = True) -> DataLoader:

        tensors = []
        sequences = [(sequence_id, dataset.instances[PLACEHOLDER_FIELD][sequence_id]) for sequence_id in
                     dataset.dataframe[dataset.instances_ids_field]]
        
        esm_callable = ESM_FUNCTIONS[self.model_name]

        _, alphabet = esm_callable()
        batch_converter = alphabet.get_batch_converter()

        # _, _, tokens = batch_converter(sequences)

        batch_size = 10000  # You can adjust this based on your preferences

        # Initialize the progress bar
        progress_bar = tqdm(total=len(sequences), desc="Processing sequences", position=0, leave=True)

        # Define the function to be parallelized
        def process_batch(batch):
            _, _, tokens = batch_converter(batch)
            return tokens

        # Process sequences in parallel with a progress bar in batches
        result_tokens = []
        for i in range(0, len(sequences), batch_size):
            batch = sequences[i:i + batch_size]
            batch_results = Parallel(n_jobs=-1)(delayed(process_batch)(batch) for batch in [batch])
            result_tokens.extend(batch_results)
            progress_bar.update(len(batch))

        # Close the progress bar
        progress_bar.close()

        # Use joblib to parallelize the function across sequences
        result_tokens = torch.cat(result_tokens, dim=0)
        tensors.append(result_tokens)

        try:
            if dataset.y is not None:
                tensors.append(torch.tensor(dataset.y, dtype=torch.float))
        except ValueError:
            pass

        dataset = TensorDataset(
            *tensors
        )
        data_loader = DataLoader(
            dataset,
            shuffle=shuffle,
            batch_size=self.batch_size
        )
        return data_loader
    


from collections import ChainMap
from typing import Callable

from joblib import Parallel, delayed

from plants_sm.data_structures.dataset import Dataset

import contextlib
import joblib
from tqdm import tqdm


@contextlib.contextmanager
def tqdm_joblib(tqdm_object):
    """Context manager to patch joblib to report into tqdm progress bar given as argument"""

    class TqdmBatchCompletionCallback(joblib.parallel.BatchCompletionCallBack):
        def __call__(self, *args, **kwargs):
            tqdm_object.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallback
    try:
        yield tqdm_object
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_object.close()


def transform_instances(n_jobs: int, dataset: Dataset, func: Callable, instance_type: str, name: str) -> Dataset:
    """
    Transform instances in a dataset

    Parameters
    ----------
    n_jobs: int
       number of jobs to run in parallel
    dataset: Dataset
        dataset to be transformed where instances are the representation or object to be processed.
    func: Callable
        function to be applied to each instance
    instance_type: str
        type of instance to be transformed
    name: str
        name of the transformation

    Returns
    -------
    dataset with transformed instances: Dataset
    """
    parallel_callback = Parallel(n_jobs=n_jobs, prefer="threads")
    instances = dataset.get_instances(instance_type)
    with tqdm_joblib(tqdm(desc=name, total=len(instances.items()))):
        res = parallel_callback(
            delayed(func)(instance_representation, instance_id)
            for instance_id, instance_representation in instances.items())

    sequences_dict = dict(res)
    dataset.instances[instance_type] = sequences_dict

    return dataset



import numpy as np

from plants_sm.data_structures.dataset import Dataset
from plants_sm.featurization.featurizer import FeaturesGenerator
from plants_sm.featurization.proteins.encodings._constants import NLF


class NLFEncoder(FeaturesGenerator):
    """
    It encodes protein sequences using the Fisher Transform.
    The ProteinNLF operates only over pandas DataFrame.

    Method that takes many physicochemical properties and transforms them using a Fisher Transform (similar to a PCA)
    creating a smaller set of features that can describe the amino acid just as well.
    There are 19 transformed features.
    This method of encoding is detailed by Nanni and Lumini in their paper:
    L. Nanni and A. Lumini, “A new encoding technique for peptide classification,”
    Expert Syst. Appl., vol. 38, no. 4, pp. 3185–3191, 2011
    This function just receives 20aa letters, therefore preprocessing is required.
    """
    name = 'NLFEncoder'
    encoding_dimension = 18
    output_shape_dimension = 3

    def set_features_names(self):
        """
        Set the features names.
        """
        self.features_names = [f"{self.name}_{num}" for num in range(1, self.encoding_dimension + 1)]

    def _fit(self, dataset: Dataset, instance_type: str) -> 'NLFEncoder':
        """
        Fit the transformer to the data.

        Parameters
        ----------
        dataset: Dataset
            The input data.

        Returns
        -------
        NLFEncoder: The fitted transformer.

        """
        return self

    def _fit_batch(self, dataset: Dataset, instance_type: str) -> 'NLFEncoder':
        """
        Fit the transformer to the data.

        Parameters
        ----------
        dataset: Dataset
            The input data.

        Returns
        -------
        NLFEncoder: The fitted transformer.

        """
        return self

    def _featurize(self, sequence: str) -> np.ndarray:
        """
        Encode a single sequence.

        Parameters
        ----------
        sequence: str
            The input sequence.

        Returns
        -------
        np.ndarray: The encoded sequence.

        """
        return np.array([np.array(NLF[aa]) for aa in sequence], dtype=np.float32)



from abc import ABCMeta, abstractmethod
from typing import Any, List

from plants_sm.io.commons import FilePathOrBuffer, get_buffer


class Reader(metaclass=ABCMeta):
    """
    Abstract class that aims at reading files of any format
    """

    def __init__(self, file_path_or_buffer: FilePathOrBuffer, **kwargs):
        """
        Constructor

        Parameters
        ----------
        file_path_or_buffer: FilePathOrBuffer

        """
        self.path = file_path_or_buffer
        if "get_buffer" in kwargs:
            del kwargs["get_buffer"]
        else:
            if "mode" in kwargs:
                self.buffer = get_buffer(file_path_or_buffer, mode=kwargs["mode"])
                del kwargs["mode"]
            else:
                self.buffer = get_buffer(file_path_or_buffer)
        self.kwargs = kwargs

    def close_buffer(self):
        """
        Method to close buffer.
        """
        self.buffer.close()

    @staticmethod
    @abstractmethod
    def file_types() -> List[str]:
        """
        Abstract method and property that returns the file types that the reader can read.

        Returns
        -------
        file_types : List[str]
            the file types that the reader can read.
        """

    @abstractmethod
    def read(self) -> Any:
        """
        Abstract method that aims at reading a file.

        Returns
        -------
        Any object with the information contained in the file.
        """

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_buffer()
        return False



from typing import Tuple

import numpy as np

from plants_sm.data_structures.dataset import Dataset
from plants_sm.transformation.transformer import Transformer


class XPadder(Transformer):
    padding_dimensions: Tuple

    @staticmethod
    def _pad_array_with_zeros(array, padding_config: Tuple) -> np.ndarray:
        padding_config_ = []
        for i in range(len(array.shape)):
            if padding_config[i] < array.shape[i]:
                padding_config_.append((0, 0))
            else:
                padding_config_.append((0, padding_config[i] - array.shape[i]))

        array = np.pad(array, tuple(padding_config_), mode='constant', constant_values=0)
        return array

    def _transform(self, dataset: Dataset, instance_type: str) -> Dataset:
        padded_dict = {key: self._pad_array_with_zeros(value, self.padding_dimensions)
                       for key, value in dataset.features[instance_type].items()}

        dataset.add_features(instance_type, padded_dict)
        return dataset

    def _fit(self, dataset: Dataset, instance_type: str) -> 'XPadder':
        return self

    def _fit_batch(self, dataset: Dataset, instance_type: str) -> 'XPadder':
        return self



from abc import ABCMeta, abstractmethod


class UnsupervisedModel(metaclass=ABCMeta):

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self._fitted = None

    def fit(self, x):
        self._fit(x)
        return self

    def transform(self, x):
        return self._transform(x)

    def predict(self, x):
        return self._predict(x)

    def fit_transform(self, x):
        return self.fit(x).transform(x)

    def fit_predict(self, x):
        return self.fit(x).predict(x)

    @property
    def fitted(self):
        return self._fitted

    @fitted.setter
    def fitted(self, value: bool):
        self._fitted = value

    @abstractmethod
    def _fit(self, x):
        pass

    @abstractmethod
    def _transform(self, x):
        pass

    @abstractmethod
    def _predict(self, x):
        pass




from nni.experiment import Experiment

from plants_sm.data_structures.dataset import Dataset


class NNISearcher:

    def __init__(self, configuration: dict, models: dict, search_space: dict):
        self.experiment = Experiment('local')
        self.config = configuration
        self._search_space = search_space
        if 'tuner' not in self.config:
            self.config['tuner'] = 'Random'
        self._tuner = self.config['tuner']
        if 'assessor' in self.config:
            self._assessor = self.config['assessor']
        if "max_trial_number" in self.config:
            self.experiment.config.max_trial_number = self.config['max_trial_number']
        else:
            self.experiment.config.max_trial_number = 10
        if "max_experiment_duration" in self.config:
            self.experiment.config.max_experiment_duration = self.config['max_experiment_duration']
        if "trial_concurrency" in self.config:
            self.experiment.config.trial_concurrency = self.config['trial_concurrency']
        self.experiment.config.tuner.name = self._tuner
        self.experiment.config.assessor.name = self._assessor
        self._number_of_trials = self.config['number_of_trials']
        self._models = models

    def search(self, train_dataset: Dataset, validation_dataset: Dataset = None):

        # TODO: the idea is to pickle everything into a temporary folder and then unpickle it in the trial

        for model in self._models:
            search_space = self._search_space[model]
            self.experiment.config.search_space = search_space
            self.experiment.config.trial_command = 'python3 -m plants_sm.hyperparameter_optimization.trial' \
                                                   ' --model {} --train_dataset {} --validation_dataset {}'

            self.experiment.start(8080)




import warnings
from typing import Dict, Any, Callable, Union, Tuple, List

import numpy as np
import pandas as pd
from numpy import ndarray

from plants_sm.data_structures.dataset import Dataset
from plants_sm.featurization.featurizer import FeaturesGenerator
from jax import vmap
from functools import partial
from jax_unirep.utils import get_embeddings

from jax_unirep.utils import load_params
from jax_unirep.layers import mLSTM
from jax_unirep.utils import validate_mLSTM_params

from plants_sm.featurization.proteins.bio_embeddings._utils import reduce_per_protein


class UniRepEmbeddings(FeaturesGenerator):
    """UniRep Embedder

    Alley, E.C., Khimulya, G., Biswas, S. et al. Unified rational protein
    engineering with sequence-based deep representation learning. Nat Methods
    16, 1315–1322 (2019). https://doi.org/10.1038/s41592-019-0598-1

    We use a reimplementation of unirep:

    Ma, Eric, and Arkadij Kummer. "Reimplementing Unirep in JAX." bioRxiv (2020).
    https://doi.org/10.1101/2020.05.11.088344
    """

    # An integer representing the size of the embedding.
    embedding_dimension = 1900
    # An integer representing the number of layers from the RAW output of the LM.
    number_of_layers = 1
    # dimension of the final embedding
    output_shape_dimension: int = 2

    _params: Dict[str, Any]
    _apply_fun: Callable

    def set_features_names(self):
        """
        The method features_names will set the names of the features
        """
        self.features_names = [f"unirep_{num}" for num in range(1, self.embedding_dimension + 1)]

    def _fit(self, dataset: Dataset, instance_type: str) -> 'UniRepEmbeddings':
        """
        Load the parameters of the mLSTM model.
        """
        self.n_jobs = 1
        self._params = load_params()[1]
        _, self._apply_fun = mLSTM(output_dim=self.embedding_dimension)
        validate_mLSTM_params(self._params, n_outputs=self.embedding_dimension)

        if self.device:
            warnings.warn("It will run on CPU, because UniRep does not allow configuring the device")

        return self

    def _fit_batch(self, dataset: Dataset, instance_type: str) -> 'UniRepEmbeddings':
        """
        Fit the UniRep model to the dataset.

        Parameters
        ----------
        dataset: Dataset
            dataset to fit the transformer where instances are the representation or object to be processed.
        instance_type: str
            type of the instances to be featurized

        Returns
        -------
        self: Estimator
            the fitted UniRep
        """
        return self._fit(dataset, instance_type)

    def _featurize(self, sequence: str) -> ndarray:
        """
        Featurize a sequence using UniRep embedding.

        Parameters
        ----------
        sequence: str
            A string representing the sequence of a protein.

        Returns
        -------
        features_df: pd.DataFrame
            A DataFrame containing the UniRep embedding of the sequence.
        """
        # https://github.com/sacdallago/bio_embeddings/issues/117
        # Unirep only allows batching with sequences of the same length, so we don't do batching at all
        embedded_seqs = get_embeddings([sequence])
        # h and c refer to hidden and cell state
        # h contains all the hidden states, while h_final and c_final contain only the last state
        h_final, c_final, h = vmap(partial(self._apply_fun, self._params))(
            embedded_seqs
        )
        # Go from a batch of 1, which is `(1, len(sequence), 1900)`, to `len(sequence), 1900)`
        if self.output_shape_dimension <= 2:
            embedding = reduce_per_protein(h[0])
        else:
            embedding = np.asarray(h[0])
        return embedding



import os
from enum import Enum
from typing import Type

from plants_sm.models.constants import FileConstants
from plants_sm.models.model import Model
from plants_sm.models.pytorch_model import PyTorchModel
from plants_sm.models.tensorflow_model import TensorflowModel


class ModelFilesClasses(Enum):
    """Enumeration of model types."""

    PYTORCH_MODEL_WEIGHTS = PyTorchModel
    PYTORCH_MODEL_PKL = PyTorchModel

    TENSORFLOW_MODEL = TensorflowModel

    MODEL_PARAMETERS_PKL = None


class ModelFileEnumeratorsUtils:
    """Class with utilities for model file enumerators."""

    @staticmethod
    def get_class_to_load_from_directory(path: str) -> Type[Model]:
        """
        Get the class to load from the directory.

        Parameters
        ----------
        path: str
            path to the directory

        Returns
        -------
        class: Type[Model]
            class to load
        """

        files = os.listdir(path)
        for file in files:
            for enumerator, model_file in FileConstants.__members__.items():
                if file == model_file.value:
                    if ModelFilesClasses[enumerator].value:
                        return ModelFilesClasses[enumerator].value





import pandas as pd
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord


def convert_csv_to_fasta(csv_file, sequence_column, header_column, output_file):
    df = pd.read_csv(csv_file)

    records = []
    for index, row in df.iterrows():
        sequence = row[sequence_column]
        header = row[header_column] if header_column else f"sequence_{index}"
        record = SeqRecord(Seq(sequence), id=header, description="")
        records.append(record)

    with open(output_file, 'w') as output_handle:
        SeqIO.write(records, output_handle, "fasta")


from collections import Counter
from itertools import product

from ._constants import AA_ALPHABET
from ._utils import Descriptor


class AminoAcidCompositionDescriptor(Descriptor):
    """
    A descriptor that returns the Amino Acid Composition of the sequence.
    """

    def __call__(self, sequence: str, **kwargs):
        """
        Calculates the Amino Acid Composition of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the isoelectric point for.

        kwargs

        Returns
        -------
        list of float
            The Amino Acid Composition of the sequence.
        """
        counts = Counter(sequence)
        return [round(counts.get(aa, 0) / len(sequence) * 100, 3) for aa in AA_ALPHABET]

    def get_features_out(self, **kwargs):
        return list(AA_ALPHABET)


class DipeptideCompositionDescriptor(Descriptor):
    """
    A descriptor that returns the Dipeptide Composition of the sequence.
    """

    def __call__(self, sequence: str, **kwargs):
        """
        Calculates the Dipeptide Composition of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the Dipeptide Composition for.

        kwargs

        Returns
        -------
        list of float
            The Dipeptide Composition of the sequence.
        """
        counts = {a + b: 0 for a, b in product(AA_ALPHABET, AA_ALPHABET)}

        for a, b in zip(sequence, sequence[1:]):
            counts[a + b] += 1

        return [round(cnt / (len(sequence) - 1) * 100, 3) for cnt in counts.values()]

    def get_features_out(self, **kwargs):
        return [f'{aa_1}{aa_2}' for aa_1, aa_2 in product(AA_ALPHABET, AA_ALPHABET)]


class TripeptideCompositionDescriptor(Descriptor):
    """
    A descriptor that returns the Tripeptide Composition of the sequence.
    """

    def __call__(self, sequence: str, **kwargs):
        """
        Calculates the Tripeptide Composition of the sequence.

        Parameters
        ----------
        sequence : str
            The sequence to calculate the Tripeptide Composition for.

        kwargs

        Returns
        -------
        list of float
            The Tripeptide Composition of the sequence.
        """
        counts = {(a, b, c): 0 for a, b, c in product(AA_ALPHABET, AA_ALPHABET, AA_ALPHABET)}

        for a, b, c in zip(sequence, sequence[1:], sequence[2:]):
            counts[(a, b, c)] += 1

        return [round(cnt / (len(sequence) - 2) * 100, 3) for cnt in counts.values()]

    def get_features_out(self, **kwargs):
        return [f'{aa_1}{aa_2}{aa_3}' for aa_1, aa_2, aa_3 in product(AA_ALPHABET, AA_ALPHABET, AA_ALPHABET)]



import json
import os

import numpy as np
from matplotlib import pyplot, pyplot as plt, cm
from sklearn.metrics import accuracy_score, balanced_accuracy_score, roc_curve, \
    roc_auc_score, precision_score, recall_score, f1_score, matthews_corrcoef, confusion_matrix, average_precision_score

import seaborn as sns

from plants_sm.data_structures.dataset import Dataset
from plants_sm.models.constants import BINARY
from plants_sm.models.model import Model


class ModelReport:

    def __init__(self, model: Model, task_type: str, dataset: Dataset, reports_directory="./"):
        """
        Constructor for the ModelReport class.

        Parameters
        ----------
        model: Model
            model to generate report for
        task_type: str
            Type of task, e.g. BINARY, MULTICLASS, REGRESSION
        dataset:
            Dataset to evaluate the model on
        reports_directory:
            Directory to save the reports to
        """
        self.model = model
        self.task_type = task_type
        self.dataset = dataset
        os.makedirs(reports_directory, exist_ok=True)
        self.reports_directory = reports_directory

    def generate_metrics_report(self):
        """
        Generates a report for the model.

        """
        if self.task_type == BINARY:
            self._generate_binary_classification_report()

    def _generate_confusion_matrix(self, confusion_matrix_data):
        """
        Generates a confusion matrix for the model.

        Parameters
        ----------
        confusion_matrix_data: np.ndarray
            Confusion matrix data

        """
        ax = plt.subplot()
        sns.heatmap(confusion_matrix_data, annot=True, fmt='g',
                    ax=ax)  # annot=True to annotate cells, ftm='g' to disable scientific notation

        # labels, title and ticks
        ax.set_xlabel('Predicted labels')
        ax.set_ylabel('True labels')
        ax.set_title('Confusion Matrix')
        ax.xaxis.set_ticklabels(['Negative', 'Positive'])
        ax.yaxis.set_ticklabels(['Negative', 'Positive'])
        pyplot.savefig(os.path.join(self.reports_directory, "confusion_matrix.png"))
        plt.show()

    def _generate_roc_auc_curve(self):
        """
        Generates a ROC AUC curve for the model.

        """
        ns_probs = [0 for _ in range(len(self.dataset.y))]
        probs = self.model.predict_proba(self.dataset)
        # keep probabilities for the positive outcome only
        # calculate scores
        # summarize scores
        ns_fpr, ns_tpr, _ = roc_curve(self.dataset.y, ns_probs)
        lr_fpr, lr_tpr, _ = roc_curve(self.dataset.y, probs)
        # plot the roc curve for the model
        pyplot.plot(ns_fpr, ns_tpr, linestyle='--', label='No Skill')
        pyplot.plot(lr_fpr, lr_tpr, marker='.', label='Model')
        # axis labels
        pyplot.xlabel('False Positive Rate')
        pyplot.ylabel('True Positive Rate')
        # show the legend
        pyplot.legend()
        # show the plot
        pyplot.savefig(os.path.join(self.reports_directory, "roc_curve.png"))
        pyplot.show()

    def _generate_history(self):
        """
        Generates a history plot for the model.
        """
        ax2 = plt.subplot()
        sns.lineplot(self.model.history["loss"], ax=ax2)
        pyplot.savefig(os.path.join(self.reports_directory, "loss_history.png"))
        plt.show()

        ax2 = plt.subplot()
        sns.lineplot(self.model.history["metric_results"], ax=ax2)
        pyplot.savefig(os.path.join(self.reports_directory, "metric_results_history.png"))
        plt.show()

    def _generate_precision_recall_curve(self, predictions_proba):
        """
        Generates a precision recall curve for the model.

        """
        probability_thresholds = np.linspace(0, 1, num=100)
        precision_scores = []
        recall_scores = []

        # Find true positive / false positive rate for each threshold
        for p in probability_thresholds:

            y_test_preds = []

            for prob in predictions_proba:
                if prob > p:
                    y_test_preds.append(1)
                else:
                    y_test_preds.append(0)

            precision = precision_score(self.dataset.y, y_test_preds)
            recall = recall_score(self.dataset.y, y_test_preds)

            precision_scores.append(precision)
            recall_scores.append(recall)

        fig, ax = plt.subplots(figsize=(6, 6))
        ax.plot(recall_scores, precision_scores, label='Model')
        baseline = len(self.dataset.y[self.dataset.y == 1]) / len(self.dataset.y)
        ax.plot([0, 1], [baseline, baseline], linestyle='--', label='Baseline')
        ax.set_xlabel('Recall')
        ax.set_ylabel('Precision')
        ax.legend(loc='center left')
        ax.set_title('Precision-Recall Curve')
        pyplot.savefig(os.path.join(self.reports_directory, "precision_recall_curve.png"))
        plt.show()

    def _generate_binary_classification_report(self):
        """
        Generates a report for a binary classification model.
        """
        predictions = self.model.predict(self.dataset)
        accuracy = accuracy_score(self.dataset.y, predictions)
        precision = precision_score(self.dataset.y, predictions)
        recall = recall_score(self.dataset.y, predictions)
        f1 = f1_score(self.dataset.y, predictions)
        balanced_accuracy = balanced_accuracy_score(self.dataset.y, predictions)
        mcc = matthews_corrcoef(self.dataset.y, predictions)
        confusion_matrix_data = confusion_matrix(self.dataset.y, predictions)
        predictions_proba = self.model.predict_proba(self.dataset)
        roc_auc = roc_auc_score(self.dataset.y, predictions_proba)
        precision_recall_auc = average_precision_score(self.dataset.y, predictions_proba)

        metrics = {
            "accuracy": accuracy,
            "precision": precision,
            "recall": recall,
            "f1": f1,
            "balanced accuracy": balanced_accuracy,
            "roc_auc": roc_auc,
            "mcc": mcc,
            "precision_recall_auc": precision_recall_auc
        }

        self.create_visual(metrics)
        pyplot.show()

        self._generate_confusion_matrix(confusion_matrix_data)
        json.dump(metrics, open(os.path.join(self.reports_directory, "metrics.json"), "w"))

        self._generate_roc_auc_curve()
        self._generate_history()

        self._generate_precision_recall_curve(predictions_proba)

    def create_visual(self, metrics):
        colors = []
        colormap = plt.get_cmap('Blues')

        for values in metrics.values():
            # z_score = (y_val - mean) / stds[n]
            # p_value = st.norm.cdf(z_score)
            colors.append(colormap(values))

        cbar = plt.colorbar(cm.ScalarMappable(norm=cm.colors.Normalize(),
                                              cmap=colormap),
                            orientation='horizontal',
                            shrink=0.5,
                            pad=0.0625,
                            ax=plt.gca())
        for l in cbar.ax.xaxis.get_ticklabels():
            l.set_fontsize(8)
        cbar.ax.tick_params(length=0)
        cbar.outline.set_linewidth(0.25)

        xAxis = [key for key, value in metrics.items()]
        yAxis = [value for key, value in metrics.items()]

        plt.bar(xAxis,
                yAxis,
                edgecolor='k',
                lw=.25,
                color=colors,
                width=.5)
        plt.grid(False, zorder=0)

        plt.title('Model metrics',
                  fontsize=10,
                  alpha=0.8)

        i = list(np.arange(len(xAxis)))
        plt.xticks(i,
                   tuple(list(metrics.keys())))

        plt.ylim([0, 1])
        # plt.xlim([-0.5, 3.5])
        plt.gca().tick_params(length=0)

        plt.xticks(fontsize=8,
                   alpha=0.8, rotation=45)
        plt.yticks(fontsize=8,
                   alpha=0.8)

        for spine in plt.gca().spines.values():
            spine.set_visible(False)

        plt.gca().tick_params(length=0)
        pyplot.savefig(os.path.join(self.reports_directory, "metrics.png"))



import os
import uuid
from abc import ABCMeta, abstractmethod

import numpy as np

from plants_sm.data_structures.dataset import Dataset
from plants_sm.io.json import write_json


class Model(metaclass=ABCMeta):

    _name: str = None

    @property
    def name(self) -> str:
        if self._name is None:
            value = str(uuid.uuid4().hex)
            self._name = value
        return self._name

    @name.setter
    def name(self, value: str):
        """
        Set the name of the model

        Parameters
        ----------
        value: str
            Name of the model
        """
        if value is None:
            value = str(uuid.uuid4().hex)
        self._name = value

    @abstractmethod
    def _preprocess_data(self, dataset: Dataset, **kwargs) -> Dataset:
        """
        Preprocesses the data.

        Parameters
        ----------
        dataset: Dataset
            The dataset to preprocess.
        kwargs:
            Additional keyword arguments.

        Returns
        -------
        Dataset
            The preprocessed dataset.
        """

    @abstractmethod
    def _fit_data(self, train_dataset: Dataset, validation_dataset: Dataset, **kwargs):
        """
        Fits the model to the data.

        Parameters
        ----------
        train_dataset: Dataset
            The dataset to fit the model to.
        validation_dataset: Dataset
            The dataset to validate the model on.
        """

    @abstractmethod
    def _predict_proba(self, dataset: Dataset, **kwargs) -> np.ndarray:
        """
        Predicts the probabilities of the classes.

        Parameters
        ----------
        dataset: Dataset
            The dataset to predict the probabilities on.

        Returns
        -------
        np.ndarray
            The predicted probabilities.
        """

    @abstractmethod
    def _predict(self, dataset: Dataset, **kwargs) -> np.ndarray:
        """
        Predicts the classes.

        Parameters
        ----------
        dataset: Dataset
            The dataset to predict the classes on.

        Returns
        -------
        np.ndarray
            The predicted classes.
        """

    @abstractmethod
    def _save(self, path: str):
        """
        Saves the model to a file.

        Parameters
        ----------
        path: str
            The path to save the model to.
        """
    @classmethod
    def load(cls, path: str):
        """
        Loads the model from a file.

        Parameters
        ----------
        path: str
            The path to load the model from.
        """
        return cls._load(path)

    @classmethod
    @abstractmethod
    def _load(cls, path: str):
        """
        Loads the model from a file.

        Parameters
        ----------
        path: str
            The path to load the model from.
        """

    @property
    @abstractmethod
    def history(self):
        """
        Returns the underlying model.
        """

    def fit(self, train_dataset: Dataset, validation_dataset: Dataset = None, **kwargs) -> 'Model':
        """
        Fits the model to the data.

        Parameters
        ----------
        train_dataset: Dataset
            The dataset to fit the model to.
        validation_dataset: Dataset
            The dataset to validate the model on.

        Returns
        -------
        self
        """
        return self._fit_data(train_dataset, validation_dataset, **kwargs)

    def predict_proba(self, dataset: Dataset, **kwargs) -> np.ndarray:
        """
        Predicts the probabilities of the classes.

        Parameters
        ----------
        dataset: Dataset
            The dataset to predict the probabilities on.
        Returns
        -------
        np.ndarray
            The predicted probabilities.
        """

        return self._predict_proba(dataset, **kwargs)

    def predict(self, dataset: Dataset, **kwargs) -> np.ndarray:
        """
        Predicts the classes.

        Parameters
        ----------
        dataset: Dataset
            The dataset to predict the classes on.

        Returns
        -------
        np.ndarray
            The predicted classes.
        """
        return self._predict(dataset, **kwargs)

    def preprocess(self, dataset: Dataset, **kwargs):
        """
        Preprocesses the data.

        Parameters
        ----------
        dataset: Dataset
            The dataset to preprocess.
        kwargs:
            Additional keyword arguments.

        Returns
        -------
        Dataset
            The preprocessed dataset.
        """
        return self._preprocess_data(dataset, **kwargs)

    def save(self, path: str):
        """
        Saves the model to a file.

        Parameters
        ----------
        path: str
            The path to save the model to.
        """
        if os.path.exists(path):
            if os.path.isdir(path):
                self._save(path)
            else:
                raise ValueError(f'Path {path} is not a directory.')

        else:
            os.makedirs(path)
            self._save(path)

        write_json(os.path.join(path, 'model_type.json'), {"type": self.__class__.__name__})




from typing import Union
import torch
import torch.nn as nn
from esm.model.esm2 import ESM2


class FineTuneESM2(ESM2):

    def forward(self, tokens, repr_layers=[], need_head_weights=False, return_contacts=False, layers_to_freeze=0):
        if return_contacts:
            need_head_weights = True

        assert tokens.ndim == 2
        padding_mask = tokens.eq(self.padding_idx)  # B, T

        x = self.embed_scale * self.embed_tokens(tokens)

        if self.token_dropout:
            x.masked_fill_((tokens == self.mask_idx).unsqueeze(-1), 0.0)
            # x: B x T x C
            mask_ratio_train = 0.15 * 0.8
            src_lengths = (~padding_mask).sum(-1)
            mask_ratio_observed = (tokens == self.mask_idx).sum(-1).to(x.dtype) / src_lengths
            x = x * (1 - mask_ratio_train) / (1 - mask_ratio_observed)[:, None, None]

        if padding_mask is not None:
            x = x * (1 - padding_mask.unsqueeze(-1).type_as(x))

        repr_layers = set(repr_layers)
        hidden_representations = {}
        if 0 in repr_layers:
            hidden_representations[0] = x

        if need_head_weights:
            attn_weights = []

        # (B, T, E) => (T, B, E)
        x = x.transpose(0, 1)

        if not padding_mask.any():
            padding_mask = None

        for layer_idx, layer in enumerate(self.layers):
            if layer_idx + 1 <= layers_to_freeze:
                with torch.no_grad():
                    x, attn = layer(
                        x,
                        self_attn_padding_mask=padding_mask,
                        need_head_weights=need_head_weights,
                    )
            else:

                x, attn = layer(
                    x,
                    self_attn_padding_mask=padding_mask,
                    need_head_weights=need_head_weights,
                )
            if (layer_idx + 1) in repr_layers:
                hidden_representations[layer_idx + 1] = x.transpose(0, 1)
            if need_head_weights:
                # (H, B, T, T) => (B, H, T, T)
                attn_weights.append(attn.transpose(1, 0))

        x = self.emb_layer_norm_after(x)
        x = x.transpose(0, 1)  # (T, B, E) => (B, T, E)

        # last hidden representation should have layer norm applied
        if (layer_idx + 1) in repr_layers:
            hidden_representations[layer_idx + 1] = x

        # with torch.no_grad():
        #     x = self.lm_head(x)

        result = {"logits": x, "representations": hidden_representations}
        if need_head_weights:
            # attentions: B x L x H x T x T
            attentions = torch.stack(attn_weights, 1)
            if padding_mask is not None:
                attention_mask = 1 - padding_mask.type_as(attentions)
                attention_mask = attention_mask.unsqueeze(1) * attention_mask.unsqueeze(2)
                attentions = attentions * attention_mask[:, None, None, :, :]
            result["attentions"] = attentions
            # if return_contacts:
            #     contacts = self.contact_head(tokens, attentions)
            #     result["contacts"] = contacts

        return result


from copy import copy

import torch
from torch import nn
from torch.nn.functional import elu, sigmoid

import torch.nn.functional as F


class DeepConvDTI(nn.Module):

    def __init__(self, protein_strides, proteins_layers, compounds_layers, interaction_layers, filters):
        super().__init__()
        self.proteins_embedding = nn.Embedding(num_embeddings=21, embedding_dim=20)
        self.protein_strides = protein_strides
        self.filters = filters
        self.proteins_layers = proteins_layers
        self.compounds_layers = compounds_layers
        self.interaction_layers = interaction_layers

    class PLayer(nn.Module):

        def __init__(self, in_channels, out_channels, stride):
            super().__init__()
            self.conv = nn.Conv1d(in_channels, out_channels, stride, stride=stride)
            self.bn = nn.BatchNorm1d(out_channels)
            self.elu = nn.ELU()

        def forward(self, x):
            x = self.conv.cuda()(x)
            x = F.normalize(x, p=2, dim=1)
            x = self.bn.cuda()(x)
            x = self.elu(x)
            x = nn.AdaptiveMaxPool1d(2)(x)
            return x

    def forward(self, x):
        proteins = x[0]
        compounds = x[1]

        proteins = proteins.to(torch.int64)
        proteins = self.proteins_embedding(proteins)

        proteins = F.normalize(proteins, p=2, dim=1)

        proteins = nn.BatchNorm1d(proteins.shape[1]).to("cuda")(proteins)
        proteins = nn.Dropout1d(0.2)(proteins)
        conv_results = []
        conv_proteins = copy(proteins)
        for i, stride in enumerate(self.protein_strides):
            conv_results.append(self.PLayer(proteins.shape[1], self.filters, stride)(conv_proteins))

        proteins = torch.cat(conv_results, dim=1)
        proteins = proteins.view(proteins.shape[0], -1)
        for i, layer in enumerate(self.proteins_layers):
            proteins = nn.Linear(proteins.shape[1], layer).cuda()(proteins)
            proteins = F.normalize(proteins, p=2, dim=1)
            proteins = nn.BatchNorm1d(proteins.shape[1]).cuda()(proteins)
            proteins = elu(proteins).cuda()

        for i, layer in enumerate(self.compounds_layers):
            compounds = nn.Linear(compounds.shape[1], layer).cuda()(compounds)
            compounds = F.normalize(compounds, p=2, dim=1)
            compounds = nn.BatchNorm1d(compounds.shape[1]).cuda()(compounds)
            compounds = elu(compounds).cuda()

        interaction = torch.cat([proteins, compounds], dim=1)
        for i, layer in enumerate(self.interaction_layers):
            interaction = nn.Linear(interaction.shape[1], layer).cuda()(interaction)
            interaction = F.normalize(interaction, p=2, dim=1)
            interaction = nn.BatchNorm1d(interaction.shape[1]).cuda()(interaction)
            interaction = elu(interaction).cuda()

        yhat = sigmoid(nn.Linear(interaction.shape[1], 1).cuda()(interaction))

        return yhat


from abc import abstractmethod
from copy import copy
from typing import Any, Union, Set, Iterable

import numpy as np

from plants_sm.data_structures.dataset import Dataset
from plants_sm.featurization.featurizer import FeaturesGenerator
from plants_sm.tokenisation.tokeniser import Tokenizer


class Encoder(FeaturesGenerator):
    """
    Abstract method that has to be implemented by all encoders to set the tokens.

    Parameters
    ----------

    alphabet: Set[str]
        alphabet of the dataset
    tokenizer: Tokenizer
        tokenizer used to tokenize the sequences
    max_length: int
        maximum length of the sequences

    """
    alphabet: Union[Set[str], str] = set()
    tokenizer: Tokenizer = None
    max_length: int = None
    tokens: dict = {}
    _max_length_set: bool = False
    _alphabet_set: bool = False

    def __init__(self, alphabet: Union[Set[str], str] = None, tokenizer: Tokenizer = None, max_length: int = None, padding=""):
        """
        Abstract method that has to be implemented by all encoders to set the tokens.

        Parameters
        ----------

        alphabet: Set[str]
            alphabet of the dataset
        tokenizer: Tokenizer
            tokenizer used to tokenize the sequences
        max_length: int
            maximum length of the sequences
        """
        super().__init__()
        if alphabet is None:
            alphabet = set()
        else:
            self._alphabet_set = True

        self.alphabet = alphabet
        self.tokenizer = tokenizer

        if max_length is not None:
            self._max_length_set = True

        self.max_length = max_length

        self.padding = padding

    def set_features_names(self):
        """
        Set the features names of the encoded sequence.
        """
        for i, token in enumerate(set(self.alphabet)):
            self.features_names.append(token)

    def _fit_batch(self, dataset: Dataset, instance_type: str) -> 'Encoder':
        """
        Fit the Encoder with the alphabet of the dataset.

        Parameters
        ----------
        dataset: Dataset
            dataset to fit the transformer where instances are the representation or object to be processed.

        Returns
        -------
        self: OneHotEncoder
            fitted OneHotEncoder
        """

        self.tokens = {}
        lengths = []
        if len(self.alphabet) == 0:
            sequences = list(dataset.get_instances(instance_type).values())
            for sequence in sequences:
                if self.tokenizer:
                    tokenized_sequence = self.tokenizer.tokenize(sequence)
                else:
                    tokenized_sequence = copy(sequence)
                lengths.append(len(tokenized_sequence))
                for char in tokenized_sequence:
                    if char != self.padding:
                        self.alphabet.add(char)
        else:
            if not self.max_length:
                sequences = list(dataset.get_instances(instance_type).values())
                self.max_length = max([len(sequence) for sequence in sequences])
            elif not self._max_length_set:
                sequences = list(dataset.get_instances(instance_type).values())
                max_length = max([len(sequence) for sequence in sequences])
                if max_length > self.max_length:
                    self.max_length = max_length

            if self._alphabet_set:
                if isinstance(self.alphabet, str):
                    self.alphabet = set(list(self.alphabet))
            else:
                letters = list(self.alphabet)
                for letter in letters:
                    self.alphabet.add(letter)

        if not self.max_length:
            self.max_length = max(lengths)

        for i, token in enumerate(self.alphabet):
            self._set_tokens(i, token)

        return self

    def _fit(self, dataset: Dataset, instance_type: str) -> 'Encoder':
        """
        Fit the Encoder with the alphabet of the dataset.

        Parameters
        ----------
        dataset: Dataset
            dataset to fit the transformer where instances are the representation or object to be processed.
        instance_type: str
            type of the instances to be transformed

        Returns
        -------
        self: OneHotEncoder
            fitted OneHotEncoder
        """
        lengths = []
        if len(self.alphabet) == 0:
            sequences = list(dataset.get_instances(instance_type).values())
            for sequence in sequences:
                if self.tokenizer:
                    tokenized_sequence = self.tokenizer.tokenize(sequence)
                else:
                    tokenized_sequence = copy(sequence)
                lengths.append(len(tokenized_sequence))
                for char in tokenized_sequence:
                    if char != self.padding:
                        self.alphabet.add(char)
        else:
            if not self.max_length:
                sequences = list(dataset.get_instances(instance_type).values())
                self.max_length = max([len(sequence) for sequence in sequences])
            if isinstance(self.alphabet, str):
                self.alphabet = set(list(self.alphabet))

        if not self.max_length:
            self.max_length = max(lengths)

        alphabet = sorted(list(self.alphabet))
        for i, token in enumerate(alphabet):
            self._set_tokens(i, token)

        return self

    @abstractmethod
    def _set_tokens(self, i: int, token: str):
        """
        Set the tokens of the alphabet.

        Parameters
        ----------
        i: int
            index of the token
        token: str
            token to be set
        """
        pass

    def _featurize(self, instance: str) -> np.ndarray:
        """
        Encode the sequence with the encoding.

        Parameters
        ----------
        instance: str
            sequence to be encoded

        Returns
        -------
        encoded_sequence: np.ndarray
        """
        if self.output_shape_dimension == 3:
            res = np.zeros((self.max_length, len(self.alphabet)), dtype=np.int32)
        else:
            res = np.zeros(self.max_length, dtype=np.int32)
        if self.tokenizer:
            instance = self.tokenizer.tokenize(instance)
        for i, token in enumerate(instance[:self.max_length]):
            if token in self.tokens:
                if self.output_shape_dimension == 3:
                    res[i, :] = self.tokens[token]
                elif self.output_shape_dimension == 2:
                    res[i] = self.tokens[token]
        return res



import os
from typing import List

from plants_sm.tokenisation.tokeniser import Tokenizer
from SmilesPE.pretokenizer import atomwise_tokenizer
from SmilesPE.pretokenizer import kmer_tokenizer

import codecs
from SmilesPE.tokenizer import SPE_Tokenizer


class AtomLevelTokenizer(Tokenizer):
    """
    Tokenizer class that is used to tokenize SMILES sequences at the atom level.
    """

    def tokenize(self, sequence: str) -> List[str]:
        """
        Tokenize a SMILES sequence at the atom level.

        Parameters
        ----------
        sequence: str
            SMILES sequence to tokenize

        Returns
        -------
        List[str]
            tokenized SMILES sequence
        """
        tokens = atomwise_tokenizer(sequence)

        return tokens


class KmerTokenizer(Tokenizer):
    """
    Tokenizer class that is used to tokenize SMILES sequences at the kmer level.

    Attributes
    ----------
    kmer_size: int
        size of the kmer

    """
    kmer_size: int = 4

    def tokenize(self, sequence: str) -> List[str]:
        """
        Tokenize a SMILES sequence at the kmer level.

        Parameters
        ----------
        sequence: str
            SMILES sequence to tokenize

        Returns
        -------
        List[str]
            tokenized SMILES sequence at the kmer level
        """
        tokens = kmer_tokenizer(sequence, ngram=self.kmer_size)

        return tokens


class SPETokenizer(Tokenizer):
    """
    Tokenizer class that is used to tokenize SMILES sequences with SPE_Tokenizer.
    """

    def tokenize(self, sequence: str) -> List[str]:
        """
        Tokenize a SMILES sequence with SPE_Tokenizer.
        """

        spe_vob = codecs.open(os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                           'SPE_ChEMBL.txt'))
        spe = SPE_Tokenizer(spe_vob)
        tokens = spe.tokenize(sequence)
        tokens = tokens.split(" ")
        return tokens



from typing import List, Union

import pandas as pd

from plants_sm.io.commons import FilePathOrBuffer
from plants_sm.io.reader import Reader
from plants_sm.io.writer import Writer


class ExcelReader(Reader):

    def __init__(self, filepath_or_buffer: FilePathOrBuffer, **kwargs):
        """
        Initializer of this class that defines instance variables by calling the initializer of the parent class (
        Reader).

        Parameters
        ----------
        filepath_or_buffer : str | Path | IO[AnyStr] | TextIO
            file path
        sep : string
            the separator of each column in the file

        Returns
        -------
        """
        super().__init__(filepath_or_buffer, mode="rb", **kwargs)

    @staticmethod
    def file_types() -> List[str]:
        """
        File types that the reader can read

        Returns
        -------
        file types: List[str]
            List of file types that the reader can read
        """
        return ["xlsx"]

    def read(self) -> pd.DataFrame:
        """
        Method to read the inputted file or buffer

        Returns
        -------
        dataframe: pd.DataFrame
            dataframe with the read data
        """
        return pd.read_excel(self.buffer, **self.kwargs)


class ExcelWriter(Writer):

    def __init__(self, filepath_or_buffer: FilePathOrBuffer, **kwargs):
        """
        Initializer of this class that defines instance variables by calling the initializer of the parent class (
        Writer) and the pd.DataFrame to be written.

        Parameters
        ----------
        filepath_or_buffer : str | Path | IO[AnyStr] | TextIO
            file path
        sep : string
            the separator of each column in the file

        Returns
        -------
        """

        super().__init__(filepath_or_buffer, **kwargs)

    @staticmethod
    def file_types() -> List[str]:
        """
        File types that the reader can read

        Returns
        -------
        file types: List[str]
            List of file types that the reader can read
        """
        return ["xlsx", "xls", "xlsx", "xlsm", "xlsb", "odf", "ods", "odt"]

    def write(self, object_to_be_written: pd.DataFrame) -> bool:
        """

        Parameters
        ----------
        object_to_be_written: pd.DataFrame
            pandas DataFrame to be written

        Returns
        -------
        flag : boolean
            whether the pd.DataFrame was written without errors or not
        """
        try:
            object_to_be_written.to_excel(self.path, **self.kwargs)
            return True
        except FileNotFoundError:
            return False


def read_excel(file_path: FilePathOrBuffer, **kwargs) -> pd.DataFrame:
    """
    Function to read Excel file calling the ExcelReader method.

    Parameters
    ----------
    file_path: FilePathOrBuffer
        path of the file to be read
    Returns
    -------
    data: pd.DataFrame
        pandas dataframe with the read data
    """
    reader = ExcelReader(file_path, **kwargs)
    df = reader.read()
    reader.close_buffer()
    return df


def write_excel(output_path: FilePathOrBuffer, object_to_be_written: pd.DataFrame, **kwargs) -> bool:
    """
    Function to write an Excel file calling the ExcelWriter method.

    Parameters
    ----------
    output_path: FilePathOrBuffer
        path or buffer of the file to be written
    object_to_be_written: pd.DataFrame
        pandas dataframe to be written

    Returns
    -------
    flag:bool
        flag that says whether the file was written
    """
    writer = ExcelWriter(filepath_or_buffer=output_path, **kwargs)
    correctly_written = writer.write(object_to_be_written)
    writer.close_buffer()
    return correctly_written




def process_slices(_columns, _slice):

    if _slice.start is not None:
        start = _slice.start
    else:
        start = 0

    if _slice.stop is not None and _slice.stop != -1:
        stop = _slice.stop
    else:
        stop = _columns.size

    if _slice.step is not None:
        step = _slice.step
    else:
        step = 1

    indexes_list = list(range(start, stop, step))
    return indexes_list
